#!/usr/bin/env python
import argparse
import csv
import os
import pandas as pd
from common import *

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import seaborn as sb
import numpy as np
sb.set(style="whitegrid")
        
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", default="-")
    parser.add_argument("-o", "--output")
    parser.add_argument(
        "-t", "--tool-name-file",
        help="File that maps profile names to display names for tools")
    parser.add_argument(
        "-T", "--threads",
        type=int, default=4, help="Thread value to keep.")
    parser.add_argument(
        "--exclude-discarded", 
        action="store_true", default=False)
    args = parser.parse_args()
    
    # Replace tool names with display versions
    tool_name_table = None
    if args.tool_name_file:
        tool_name_table = {}
        with open(args.tool_name_file, 'rt') as inp:
            for row in csv.DictReader(inp, delimiter="\t"):
                tool_name_table[row['ProfileName']] = row['DisplayName']
    
    txt_file = "{}.txt".format(args.output)
    
    if not os.path.exists(txt_file):
        with open(args.input, 'rt') as inp, open(txt_file, 'wt') as out:
            writer = csv.writer(out, delimiter="\t")
            writer.writerow(('Read', 'Program', 'MAPQ', 'In Region?'))
            read_name = None
            read1 = None
            read2 = None
            
            def add_entry(row, read, read_dict):
                if read is not None:
                    q = int(row['read{}_quality'.format(read)])
                    if args.exclude_discarded and q1 < 0:
                        return True
                    else:
                        entry = (
                            "{}.{}".format(read_name, read), prog, q,
                            bool(int(row['read{}_in_region'.format(read)])))
                        if prog in read_dict:
                            assert read_dict[prog] == entry
                        else:
                            read_dict[prog] = entry
                
            for row in csv.DictReader(inp, delimiter="\t"):
                # Since neither threads nor compression change the outcome of trimming,
                # we don't need to worry about stratifying by those metric.
                prog = row['prog']
                if not (prog == 'untrimmed' or int(row['threads']) == args.threads):
                    continue
                
                if read_name != row['read_name']:
                    if read_name:
                        if read1:
                            writer.writerows(read1.values())
                        if read2:
                            writer.writerows(read2.values())
                    read_name = row['read_name']
                    read1 = {}
                    read2 = {}
                
                if add_entry(row, 1, read1):
                    read1 = None
                if add_entry(row, 2, read2):
                    read2 = None
    
    with fileopen(txt_file, 'rt') as inp:
        table = pd.read_csv(inp, sep="\t")
    
    table = (table
        .groupby(['Program', 'MAPQ', 'In Region?'])
        .size()
        .reset_index())
    table = table.rename(columns={ 0 : 'Count' })
    
    untrimmed = table[
        (table.Program == 'untrimmed') & 
        (table.MAPQ >= 0) & 
        (table['In Region?'] == True)]
    trimmed = table[
        (table.Program != 'untrimmed') & 
        (table.MAPQ >= 0) & (table['In Region?'] == True)].copy()
    trimmed.Delta = np.NaN
    for MAPQ in trimmed.MAPQ.unique():
            u = untrimmed.loc[untrimmed.MAPQ==MAPQ, 'Count'].values
            for prog in trimmed.Program.unique():
                t = trimmed.loc[
                    (trimmed.Program==prog) & (trimmed.MAPQ==MAPQ), 
                    'Count'].values
                if len(t) > 0:
                    trimmed.loc[
                        (trimmed.Program==prog) & (trimmed.MAPQ==MAPQ) , 
                        'Delta'] = t[0] - u[0]
    
    plot = sb.factorplot(
        x='MAPQ', y='Delta', hue='Program', col='In Region?', 
        data=trimmed, kind="bar", ci=None, sharey=True)
    plot.set_xlabels('Mapping Quality Score (MAPQ) Cutoff')
    plot.set_ylabels('Difference versus Untrimmed Reads')
    svg_file = args.output + ".svg"
    plot.savefig(svg_file)

if __name__ == "__main__":
    main()



"""Report generator for the detect command.
"""
from itertools import repeat
from atropos.commands.reports import BaseReportGenerator
from atropos.commands.legacy_report import Printer, TitlePrinter
from atropos.io import open_output
from atropos.io.seqio import FastaFormat

class ReportGenerator(BaseReportGenerator):
    def get_report_args(self, fmt, options):
        if fmt == 'fasta':
            if options.fasta:
                return dict((opt, True) for opt in options.fasta)
            else:
                return dict(perinput=True)
        return {}
    
    def generate_text_report(self, fmt, summary, outfile, **kwargs):
        if fmt == 'txt':
            with open_output(outfile, context_wrapper=True) as out:
                generate_reports(out, summary, **kwargs)
        elif fmt == 'fasta':
            generate_fasta(outfile, summary, **kwargs)
        else:
            super().generate_from_template(fmt, summary, outfile, **kwargs)

def generate_reports(outstream, summary):
    """Prints text reports for the results from one or a pair of detectors.
    """
    names = summary['input']['input_names'] or repeat(None)
    n_reads = summary['record_counts'][0]
    for input_idx, (matches, name) in enumerate(zip(summary['detect']['matches'], names), 1):
        generate_detector_report(outstream, input_idx, n_reads, matches, name)

def generate_detector_report(outstream, input_idx, n_reads, matches, input_name=None):
    n_matches = len(matches)
    pad_size = len(str(n_matches))
    
    _print = Printer(outstream)
    _print_title = TitlePrinter(outstream)
    _print_indent = Printer(outstream, indent=' ' * (pad_size + 2))
    
    _print.newline()
    _print_title("Input {}".format(input_idx), level=0)
    
    if input_name:
        _print("File: {}".format(input_name))
    
    _print("Detected {} adapters/contaminants:".format(n_matches))
    
    if n_matches == 0:
        _print("Try increasing --max-reads")
        return
    
    for idx, match in enumerate(matches):
        _print(
            ("{:>" + str(pad_size) + "}. Longest kmer: {}").format(
                idx+1, match['longest_kmer']))
        if match['longest_match']:
            _print_indent("Longest matching sequence: {}".format(
                match['longest_match']))
        if match['is_known']:
            _print_indent("Name(s): {}".format(
                ",\n{}".format(' ' * (pad_size + 11)).join(match['known_names'])))
            _print_indent("Known sequence(s): {}".format(
                ",\n{}".format(' ' * (pad_size + 11)).join(match['known_seqs'])))
            _print_indent(
                "Known sequence K-mers that match detected contaminant: "
                "{:.2%}".format(match['known_to_contaminant_match_frac']))
        if match['abundance']:
            _print_indent("Abundance (full-length) in {} reads: {} ({:.1%})".format(
                n_reads, match['abundance'], match['abundance'] / n_reads))
        if match['contaminant_to_known_match_frac']:
            _print_indent(
                "Detected contaminant kmers that match known sequence: "
                "{:.2%}".format(match['contaminant_to_known_match_frac']))
        if match['kmer_freq_type'] == 'frequency':
            _print_indent("Frequency of k-mers: {:.2%}".format(match['kmer_freq']))
        else:
            _print_indent("Number of k-mer matches: {}".format(match['kmer_freq']))

def generate_fasta(outfile, summary, union=False, perinput=False):
    names = summary['input']['input_names'] or repeat(None)
    n_reads = summary['record_counts'][0]
    fasta_format = FastaFormat()
    if union:
        union_records = []
    if perinput:
        if outfile.endswith('.fasta'):
            name_prefix = outfile[:-6]
        elif outfile.endswith('.fa'):
            name_prefix = outfile[:-3]
        else:
            name_prefix = outfile
    
    def format_match(idx, match, records):
        name2 = [
            "kmer_freq={}".format(match['kmer_freq']),
            "kmer_freq_type={}".format(match["kmer_freq_type"])
        ]
        if match['abundance']:
            name2.append("abundance={}".format(match['abundance']))
            name2.append("abundance_frac={}".format(match['abundance'] / n_reads))
        if match['contaminant_to_known_match_frac']:
            name2.append("contaminant_to_known_match_frac={}".format(
                match["contaminant_to_known_match_frac"]))
            
        if match['is_known']:
            name = match['known_names'][0]
            name3 = []
            if len(match['known_names']) > 1:
                name3 = ["other_names={}".format('|'.join(match['known_names'][1:]))]
            if len(match['known_seqs']) > 1:
                for seq in match['known_seqs']:
                    records.append(fasta_format.format_entry(
                        "{}.{} {}".format(name, idx, ";".join(name2 + name3)),
                        seq))
            else:
                records.append(fasta_format.format_entry(
                    "{} {}".format(name, ";".join(name2 + name3)),
                    match['known_seqs'][0]))
        else:
            records.append(fasta_format.format_entry(
                "{} {}".format(idx, ";".join(name2)), 
                match['longest_kmer']))
    
    for i, (name, matches) in enumerate(zip(names, summary['detect']['matches'])):
        records = []
        for idx, match in enumerate(matches, 1):
            format_match(idx, match, records)
        if union:
            union_records.extend(records)
        if perinput:
            with open_output("{}.{}.fasta".format(name_prefix, i), 'wt') as out:
                out.write("".join(records))
    
    if union:
        with open_output(outfile, 'wt') as union_out:
            union_out.write("".join(union_records))



# coding: utf-8
from contextlib import contextmanager
from importlib import import_module
import os
import sys
import traceback
import urllib.request
from atropos.commands import get_command


@contextmanager
def redirect_stderr():
    "Send stderr to stdout. Nose doesn't capture stderr, yet."
    old_stderr = sys.stderr
    sys.stderr = sys.stdout
    yield

    sys.stderr = old_stderr


@contextmanager
def temporary_path(name):
    directory = os.path.join(os.path.dirname(__file__), 'testtmp')
    if not os.path.isdir(directory):
        os.mkdir(directory)
    path = os.path.join(directory, name)
    yield path

    if os.path.exists(path):
        os.remove(path)


def datapath(path):
    return os.path.join(os.path.dirname(__file__), 'data', path)


def cutpath(path):
    return os.path.join(os.path.dirname(__file__), 'cut', path)


def files_equal(path1, path2):
    # return os.system("diff -u {0} {1}".format(path1, path2)) == 0
    with open(path1, 'r') as i1, open(path2, 'r') as i2:
        print("<[{}]>".format(i1.read()))
        print("<[{}]>".format(i2.read()))
    from subprocess import check_output, CalledProcessError

    try:
        check_output("diff -u {0} {1}".format(path1, path2), shell=True)
        return True

    except CalledProcessError as e:
        print("Diff: <{}>".format(e.output.decode("utf-8")))
        return False


def run(
    params,
    expected,
    inpath=None,
    inpath2=None,
    qualfile=None,
    interleaved_input=False,
    interleaved_output=False,
    sra_accn=None,
):
    if type(params) is str:
        params = params.split()
    with temporary_path(expected) as tmp_fastaq:
        if sra_accn:
            params += ['-sra', sra_accn]
        elif interleaved_input:
            params += ['-l', inpath]
        elif inpath2:
            params += ['-pe1', datapath(inpath)]
            params += ['-pe2', datapath(inpath2)]
        else:
            params += ['-se', datapath(inpath)]
            if qualfile:
                params += ['-sq', datapath(qualfile)]
        if interleaved_output:
            params += ['-L', tmp_fastaq]
        else:
            params += ['-o', tmp_fastaq]  # TODO not parallelizable
        # print(params)
        command = get_command('trim')
        retcode, summary = command.execute(params)
        assert summary is not None
        assert isinstance(summary, dict)
        if 'exception' in summary and summary['exception'] is not None:
            assert retcode != 0
            err = summary['exception']
            traceback.print_exception(* err['details'])
            raise Exception("Unexpected error: {}".format(err['message']))

        else:
            assert retcode == 0
        # TODO redirect standard output
        assert os.path.exists(tmp_fastaq)
        assert files_equal(cutpath(expected), tmp_fastaq)




# TODO diff log files
def approx_equal(a, b, tol):
    return abs(a - b) <= tol


def no_internet(url="https://github.com"):
    """Test whether there's no internet connection available.
    """
    try:
        urllib.request.urlopen(url).info()
        return False

    except:
        return True


def no_import(lib):
    """Test whether a library is importable.
    """
    try:
        mod = import_module(lib)
        return mod is None

    except:
        return True



from argparse import ArgumentParser
from atropos.modifiers import AdapterCutter, InsertAdapterCutter
from atropos.adapters import AdapterParser
from atropos.seqio import open_reader

adapter_parser = AdapterParser()
A1 = adapter_parser.parse('AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG')
A2 = adapter_parser.parse('AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT')

def create_adapter():
    ac1 = AdapterCutter([A1])
    ac2 = AdapterCutter([A2])
    return lambda read1,read2: (ac1(read1), ac2(read2))

def create_insert():
    return InsertAdapterCutter(A1, A2)

def time_process(reader, aligner):
    from datetime import datetime
    start = datetime.now()
    i = run_process(reader, aligner)
    stop = datetime.now()
    t = stop-start
    print("Time to process {} read pairs: {}".format(i, t))
    return t.total_seconds()

def run_process(reader, aligner):
    for i, (read1, read2) in enumerate(reader, 1):
        read1_new, read2_new = aligner(read1, read2)
    return i

def main():
    parser = ArgumentParser()
    parser.add_argument("-a", "--aligner", choices=('adapter', 'insert'))
    parser.add_argument("-m", "--mode", choices=('time','run'), default='run')
    parser.add_argument("-o", "--outfile")
    parser.add_argument("-1", "--fastq1", default='test1.fq')
    parser.add_argument("-2", "--fastq2", default='test2.fq')
    parser.add_argument("-i", "--iters", default=10)
    args = parser.parse_args()
    
    aligner = create_adapter() if args.aligner == 'adapter' else create_insert()
    
    if args.mode == 'time':
        t = 0
        for i in range(args.iters):
            reader = open_reader(args.fastq1, args.fastq2)
            t += time_process(reader, aligner)
            reader.close()
        print("Average time: {}".format(t / args.iters))
    else:
        run_process(reader, aligner)

if __name__ == "__main__":
    main()



"""Command line interface for the qc command.
"""
from atropos.commands.cli import (
    BaseCommandParser, parse_stat_args, configure_threads, positive, int_or_str,
    writeable_file)

class CommandParser(BaseCommandParser):
    name = 'qc'
    usage = """
atropos qc -se input.fastq
atropos qc -pe1 in1.fastq -pe2 in2.fastq
"""
    description = """
Compute read-level statistics. The output is identical to running the 'trim'
command with '--stats pre'. Use this command to get an idea of the quality of
your raw data.
"""

    def add_command_options(self):
        self.parser.set_defaults(
            action='qc',
            batch_size=None)
        
        group = self.add_group("Output")
        group.add_argument(
            "-o",
            "--output",
            type=writeable_file, default="-", metavar="FILE",
            help="Write stats to file rather than stdout.")
        
        group = self.add_group(
            "Report", title="Report content and formatting options")
        group.add_argument(
            "--report-formats",
            nargs="*", choices=("txt", "json"), default=None, metavar="FORMAT",
            help="Report type(s) to generate. If multiple, '--output' "
                 "is treated as a prefix and the appropriate extensions are "
                 "appended. If unspecified, the format is guessed from the "
                 "output file extension.")
        group.add_argument(
            "--stats",
            type=parse_stat_args, default=None,
            help="Additional arguments for read statistic collection. E.g. "
                 "'pre:tiles' means to also collect tile-level statistics "
                 "(Illumina data only), and 'pre:tiles=<regexp>' means to use "
                 "the specified regular expression to extract key portions of "
                 "read names to collect the tile statistics.")
        
        group = self.add_group(
            "Parallel", title="Parallel (multi-core) options")
        group.add_argument(
            "-T",
            "--threads",
            type=positive(int, True), default=None, metavar="THREADS",
            help="Number of threads to use for read trimming. Set to 0 to use "
                 "max available threads. (Do not use multithreading)")
        group.add_argument(
            "--process-timeout",
            type=positive(int, True), default=60, metavar="SECONDS",
            help="Number of seconds process should wait before escalating "
                 "messages to ERROR level. (60)")
        group.add_argument(
            "--read-queue-size",
            type=int_or_str, default=None, metavar="SIZE",
            help="Size of queue for batches of reads to be processed. "
                 "(THREADS * 100)")
    
    def validate_command_options(self, options):
        options.report_file = options.output
        if options.threads is not None:
            threads = configure_threads(options, self.parser)
            if options.read_queue_size is None:
                options.read_queue_size = threads * 100
            elif (
                    options.read_queue_size > 0 and
                    options.read_queue_size < threads):
                self.parser.error("Read queue size must be >= than 'threads'")
        if options.batch_size is None:
            options.batch_size = 1000



#!/usr/bin/env python
"""Given rows from parse_gtime, generate a figure and a latex table of results.
"""
import argparse
from common import fileopen
import os
import sys
import pandas as pd

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", default="-")
    parser.add_argument("-o", "--output", default="-")
    parser.add_argument("-n", "--name", default="table")
    parser.add_argument("-c", "--caption", default="")
    parser.add_argument(
        "-t", "--threads", type=int, default=None,
        help="Set all rows to have the same value for the Threads column.")
    parser.add_argument(
        "-f", "--formats", choices=('txt', 'tex', 'svg', 'pickle'), nargs='*',
        default=['tex', 'svg'])
    args = parser.parse_args()

    # read raw data
    with fileopen(args.input, "rt") as inp:
        table = pd.read_csv(
            inp, sep='\t', names=(
                'Program', 'Program2', 'Threads', 'Dataset', 'Quality', 
                'DurationSecs', 'CPUPct', 'MemoryMB'),
            dtype={ 'Program' : 'category', 'Dataset' : 'category'})
    
    if args.threads:
        table.Threads = args.threads
    
    # save table (useful if input was stdin)

    if 'txt' in args.formats:
        table.to_csv(args.output + ".txt", sep="\t", index=False)

    if 'pickle' in args.formats:
        import pickle
        pickle_file = args.output + '.pickle'
        with fileopen(pickle_file, 'wb') as out:
            pickle.dump(table, out)

    # generate latex table
    if 'tex' in args.formats:
        texdat = table.melt(
            id_vars=['Program2', 'Threads', 'Dataset', 'Quality'], 
            value_vars=['DurationSecs','CPUPct','MemoryMB'])
        texdat = (texdat.
            groupby(['Threads', 'Program2', 'variable']).
            agg({ 'value' : [min, max] }).
            sort_index())
        texdat.columns = texdat.columns.droplevel()
        from mako.template import Template
        table_template = Template(filename=os.path.join(
            os.path.dirname(__file__), "performance_table.tex"))
        tex_file = args.output + ".tex"
        with fileopen(tex_file, "wt") as o:
            o.write(table_template.render(
                name=args.name, caption=args.caption, table=texdat))

    # generate figure
    if 'svg' in args.formats:
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import seaborn as sb
        sb.set(style="whitegrid")
        import numpy as np
        
        progs = list(zip(
            ('adapterremoval', 'atropos (adapter)', 'atropos (insert)', 'seqpurge', 'skewer'),
            ('AdapterRemoval', 'Atropos (adapter)', 'Atropos (insert)', 'SeqPurge', 'Skewer'),
            sb.color_palette(n_colors=5)
        ))
        pal = dict((p[1], p[2]) for p in progs)
        
        svgdat = table.melt(
            id_vars=['Program', 'Threads', 'Dataset', 'Quality'], 
            value_vars=['DurationSecs','CPUPct','MemoryMB'])
        svgdat['Program'] = svgdat['Program'].map(dict((p[0], p[1]) for p in progs))
        svgdat['Program'] = svgdat['Program'].astype('category')
        svgdat['Dataset'] = svgdat['Dataset'].astype('category')
        svgdat['variable'] = pd.Categorical(
            svgdat['variable'], categories=['DurationSecs', 'MemoryMB', 'CPUPct'])
        
        threads = svgdat.Threads.unique()
        plot = sb.factorplot(
            x='Threads', y="value", col="variable", hue="Program", 
            data=svgdat, kind="bar", sharey=False, estimator=np.mean)
        if len(threads) == 1:
            plot.set_xticklabels('')
            plot.set_xlabels('')
        else:
            plot.set_xlabels('Threads')
        plot.axes[0,0].set_ylabel('Runtime (sec)')
        plot.axes[0,1].set_ylabel('Memory (Mb)')
        plot.axes[0,2].set_ylabel('CPU (%)')
        plot.fig.subplots_adjust(wspace=0.35)
        plot.set_titles('')
        svg_file = args.output + ".svg"
        plot.savefig(svg_file)

if __name__ == "__main__":
    main()



# coding: utf-8
"""Quality trimming.
"""
# Import cythonized functions, defaulting to pure python implementations.
try:
    from ._qualtrim import quality_trim_index, nextseq_trim_index

except:
    import logging
    from atropos.util import qual2int
    
    logging.getLogger().debug("Import failed for cythonized qualtrim functions")
    
    def quality_trim_index(qualities, cutoff_front, cutoff_back, base=33):
        """Find the position at which to trim a low-quality end from a
        nucleotide sequence.
        
        Qualities are assumed to be ASCII-encoded as chr(qual + base).
        
        The algorithm is the same as the one used by BWA within the function
        'bwa_trim_read':
        - Subtract the cutoff value from all qualities.
        - Compute partial sums from all indices to the end of the sequence.
        - Trim sequence at the index at which the sum is minimal.
        """
        start = 0
        stop = max_i = len(qualities)
        
        # find trim position for 5' end
        score = 0
        max_qual = 0
        for idx in range(max_i):
            qual = qual2int(qualities[idx], base)
            score += cutoff_front - (qual - base)
            if score < 0:
                break
            if score > max_qual:
                max_qual = score
                start = idx + 1

        # same for 3' end
        max_qual = 0
        score = 0
        for idx in reversed(range(max_i)):
            qual = qual2int(qualities[idx], base)
            score += cutoff_back - (qual - base)
            if score < 0:
                break
            if score > max_qual:
                max_qual = score
                stop = idx
        
        if start >= stop:
            start, stop = 0, 0
        
        return (start, stop)
    
    def nextseq_trim_index(sequence, cutoff, base=33):
        """Variant of the above quality trimming routine that works on NextSeq
        data. With Illumina NextSeq, bases are encoded with two colors.
        'No color' (a dark cycle) usually means that a 'G' was sequenced, but
        that also occurs when sequencing falls off the end of the fragment. The
        read then contains a run of high-quality G bases in the end.

        This routine works as the one above, but counts qualities belonging to
        'G' bases as being equal to cutoff - 1.
        """
        bases = sequence.sequence
        qualities = sequence.qualities
        score = 0
        max_qual = 0
        max_i = len(qualities)
        for idx in reversed(range(max_i)):
            qual = qual2int(qualities[idx], base)
            if bases[idx] == 'G':
                qual = cutoff - 1
            score += cutoff - qual
            if score < 0:
                break
            if score > max_qual:
                max_qual = score
                max_i = idx
        return max_i



# Tests for internal components of the atropos commands
from pytest import raises
from atropos.commands.trim.multicore import OrderPreservingWriterResultHandler
from atropos.commands.trim.writers import Writers
import tempfile
import os


def test_order_preserving_writer():
    path = tempfile.mkstemp()[1]
    try:
        writers = Writers()
        handler = OrderPreservingWriterResultHandler(writers)
        handler.start(None)
        # write three batches out of order
        result2 = "result2"
        handler.write_result(2, {path: result2})
        result3 = "result3"
        handler.write_result(3, {path: result3})
        result1 = "result1"
        handler.write_result(1, {path: result1})
        handler.finish(total_batches=3)
        # check that the results are in the right order
        with open(path, 'rt') as inp:
            assert inp.read() == (result1 + result2 + result3)
    finally:
        os.remove(path)



#!/usr/bin/env python
"""Parse machine info (cat /proc/cpuinfo /proc/meminfo) and write a 
tab-delimited row to stdout with the following fields:

1. Program name
2. Total memory
3. Number of processors
4. Speed of processors
"""
import argparse
from collections import defaultdict
from common import fileopen, parse_profile
import re

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", default="-")
    parser.add_argument("-o", "--output", default="-")
    parser.add_argument("-p", "--profile", nargs='+')
    args = parser.parse_args()
    
    profile = list(parse_profile(args.profile[0]))
    if len(args.profile) > 1:
        profile.append(args.profile[1])
    
    with fileopen(args.input, 'rt') as i:
        lines = [line.strip() for line in i.readlines()]
    
    mem_matcher = re.compile("MemTotal:\s+(\d+ .*)")
    mem = None
    
    cpu_matcher = re.compile("model name\s*:\s*(.*)")
    cpus = defaultdict(int)
    
    for line in lines:
        if mem is None:
            mem_match = mem_matcher.match(line)
            if mem_match:
                mem = mem_match.group(1)
                continue
        
        cpu_match = cpu_matcher.match(line)
        if cpu_match:
            cpus[cpu_match.group(1)] += 1
    
    total_cpus = sum(cpus.values())
    cpu_str = "; ".join("{} {}".format(count, cpu) for cpu, count in cpus.items())
    
    with fileopen(args.output, "wt") as out:
        print(*profile, total_cpus, mem, cpu_str, sep="\t", file=out)

if __name__ == "__main__":
    main()


#!/usr/bin/env python
"""Given rows from a parse job script, generates tables/figures for certain
performance metrics. Each output from the parse job script is expected to have 
one header row and one metrics row. The first 5 columns are expected to be the 
task profile information.
"""
import argparse
from common import fileopen, parse_size
import csv
import os
import pandas as pd

class Metrics():
    def __init__(self, header, rows, output, formats, template_path):
        self.header = header
        self.rows = rows
        self.output = output
        self.formats = formats
        self.template_path = template_path
    
    def show(self, name, column, **kwargs):
        table = self._get_table(column)
        prefix = "{}.{}".format(self.output, name)
        
        for fmt in self.formats:
            outfile = "{}.{}".format(prefix, fmt)
            if fmt == 'txt':
                table.to_csv(outfile, sep="\t", index=False)
            elif fmt == 'pickle':
                import pickle
                with fileopen(outfile, 'wb') as out:
                    pickle.dump(table, out)
            else:
                fn = getattr(self, "{}_{}".format(name, fmt))
                fn(table, column, outfile)
        
    def mem_tex(self, table, column, outfile, name=None, caption=None):
        texdat = (table.
            drop('Program', 1).
            rename(columns={
                'Program2' : 'Program',
                column : 'Memory'
            }).
            groupby(['Dataset', 'Threads', 'Program']).
            agg({ 'Memory' : max }))
        texdat = texdat.assign(MemoryMB=round(texdat['Memory'] / 1000000, 1))
        
        from mako.template import Template
        table_template = Template(filename=os.path.join(
            self.template_path, "job_memory_table.tex"))
        with fileopen(outfile, "wt") as o:
            o.write(table_template.render(
                name=name, caption=caption, table=texdat))
    
    def mem_svg(self, table, column, outfile):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import seaborn as sb
        sb.set(style="whitegrid")
        
        svgdat = (table.
            rename(columns={ column : 'Memory' }).
            groupby(['Dataset', 'Threads', 'Program']).
            agg({ 'Memory' : max }).
            reset_index())
        svgdat = svgdat.assign(MemoryMB=svgdat['Memory'] / 1000000)
        
        threads = svgdat.Threads.unique()
        if len(threads) == 1:
            plot = sb.factorplot(
                x='Program', y='MemoryMB', col="Dataset", 
                data=svgdat, kind="bar", ci=None, sharey=True)
        else:
            plot = sb.factorplot(
                x='Threads', y='MemoryMB', col="Dataset", hue="Program", 
                data=svgdat, kind="bar", ci=None, sharey=True)

        if len(threads) == 1:
            plot = plot.set_titles('')

        plot = plot.set_xlabels('Threads')
        plot = plot.set_ylabels('Memory (MB)')
        plot = plot.set_xticklabels(rotation=90)
        plot.fig.subplots_adjust(wspace=0.35)
        plot.savefig(outfile)
    
    def _get_table(self, column, is_size=True):
        cols = list(range(5))
        cols.append(self.header.index(column))
        header = [self.header[c] for c in cols]
        rows = [
            [row[c] for c in cols]
            for row in self.rows
        ]
        if is_size:
            for row in rows:
                row[5] = parse_size(row[5])
        table = pd.DataFrame.from_records(rows, columns=header)
        table = table.rename(columns={ 
            'prog' : 'Program',
            'prog2' : 'Program2',
            'threads' : 'Threads',
            'dataset' : 'Dataset',
            'qcut' : 'Quality',
        })
        table['Threads'] = pd.to_numeric(table['Threads'])
        table['Dataset'] = pd.Categorical(table['Dataset'])
        table['Program'] = pd.Categorical(table['Program'])
        table['Program2'] = pd.Categorical(table['Program2'])
        return table

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", default="-")
    parser.add_argument("-o", "--output", default="-")
    parser.add_argument("-m", "--metric", nargs='*', default=None)
    parser.add_argument(
        "-f", "--formats", 
        choices=('txt', 'tex', 'pickle', 'svg'), nargs='*', default=['txt'])
    args = parser.parse_args()
    
    header = None
    rows = []
    
    with fileopen(args.input, 'rt') as inp:
        for i, line in enumerate(csv.reader(inp, delimiter="\t")):
            if i == 0:
                header = line
            elif i % 2 == 0:
                if len(line) == 0:
                    break
                else:
                    assert header == line
            else:
                rows.append(line)
    
    metrics = Metrics(
        header, rows, args.output, args.formats, os.path.dirname(__file__))
    
    for metric in args.metric:
        name, column = metric.split('=')
        metrics.show(name, column)

if __name__ == '__main__':
    main()



# coding: utf-8
"""
Tests write output (should it return True or False or write)
"""
from atropos.commands.trim.filters import (
    NContentFilter, DISCARD, KEEP, SingleWrapper, PairedWrapper
)
from atropos.io.seqio import Sequence


def test_ncontentfilter():
    # third parameter is True if read should be discarded
    params = [
        ('AAA', 0, KEEP),
        ('AAA', 1, KEEP),
        ('AAACCTTGGN', 1, KEEP),
        ('AAACNNNCTTGGN', 0.5, KEEP),
        ('NNNNNN', 1, DISCARD),
        ('ANAAAA', 1 / 6, KEEP),
        ('ANAAAA', 0, DISCARD),
    ]
    for seq, count, expected in params:
        filter = NContentFilter(count=count)
        _seq = Sequence('read1', seq, qualities='#' * len(seq))
        assert filter(_seq) == expected


def test_ncontentfilter_paired():
    params = [
        ('AAA', 'AAA', 0, KEEP),
        ('AAAN', 'AAA', 0, DISCARD),
        ('AAA', 'AANA', 0, DISCARD),
        ('ANAA', 'AANA', 1, KEEP),
    ]
    for seq1, seq2, count, expected in params:
        filter = NContentFilter(count=count)
        filter_legacy = SingleWrapper(filter)
        filter_both = PairedWrapper(filter)
        read1 = Sequence('read1', seq1, qualities='#' * len(seq1))
        read2 = Sequence('read1', seq2, qualities='#' * len(seq2))
        assert filter_legacy(read1, read2) == filter(read1)
        # discard entire pair if one of the reads fulfills criteria
        assert filter_both(read1, read2) == expected



"""Report generator for the error command.

TODO: move reporting functionality out of the ErrorEstimator class.
"""
from itertools import repeat
from atropos.commands.reports import BaseReportGenerator
from atropos.io import open_output
from atropos.commands.legacy_report import Printer, TitlePrinter

class ReportGenerator(BaseReportGenerator):
    def generate_text_report(self, fmt, summary, outfile, **kwargs):
        if fmt == 'txt':
            with open_output(outfile, context_wrapper=True) as out:
                generate_reports(out, summary)
        else:
            super().generate_from_template(fmt, summary, outfile, **kwargs)

def generate_reports(outstream, summary):
    names = summary['input']['input_names'] or repeat(None)
    estimates = summary['errorrate']['estimate']
    
    _print = Printer(outstream)
    _print_title = TitlePrinter(outstream)
    
    input_idx = 0
    for input_idx, (estimate, details, name) in enumerate(zip(
            estimates, summary['errorrate']['details'], names), 1):
        generate_estimator_report(
            outstream, input_idx, estimate, details, _print, _print_title, name)
    
    if input_idx > 1:
        _print.newline()
        _print_title("Overall", level=0)
        total_lens = summary['errorrate']['total_len']
        overall_err = (
            sum(err * total_len for err, total_len in zip(estimates, total_lens)) / 
            sum(total_lens))
        print("Error rate: {:.2%}".format(overall_err), file=outstream)

def generate_estimator_report(
        outstream, input_idx, estimate, details, _print, _print_title, 
        input_name=None):
    _print_indent = Printer(outstream, indent='  ')
    
    _print.newline()
    _print_title("Input {}".format(input_idx), level=0)
    
    if input_name:
        _print("File: {}".format(input_name))
    
    _print("Error rate: {:.2%}".format(estimate))
    if details:
        _print("Details:\n")
        per_read = details['per_read']
        per_cycle = details['per_cycle']
        
        _print_indent("StdErr: {:.2%}".format(per_read['standard error']))
        _print_indent("Per-cycle rates:")
        for cycle in per_cycle:
            _print_indent(
                "Cycle: {}, Error: {:.2%}, StdErr: {:.2%}".format(*cycle), 
                indent=2)



"""File compression/decompression functions.
"""
import bz2
import gzip
import io
import lzma
import os
from subprocess import Popen, PIPE

COMPRESSORS = {
    ".gz"  : gzip,
    ".bz2" : bz2,
    ".xz"  : lzma
}
"""Mapping of file extension to python compression library."""

class GzipWriter:
    """Wrapper for a process that uses the system gzip program to compress
    bytes.
    
    Args:
        path: The path of the output file.
        mode: The file open mode.
    """
    def __init__(self, path, mode='w'):
        self.name = path
        self.outfile = open(path, mode)
        self.devnull = open(os.devnull, 'w')
        self.closed = False
        try:
            # Setting close_fds to True is necessary due to
            # http://bugs.python.org/issue12786
            self.process = Popen(
                [get_program_path('gzip')], stdin=PIPE, stdout=self.outfile,
                stderr=self.devnull, close_fds=True)
        except IOError:
            self.outfile.close()
            self.devnull.close()
            raise
    
    def readable(self):
        return False
    
    def writable(self):
        return True
    
    def seekable(self):
        return False
    
    def write(self, arg):
        self.process.stdin.write(arg)
    
    def flush(self):
        self.process.stdin.flush()
    
    def close(self):
        self.closed = True
        self.process.stdin.close()
        retcode = self.process.wait()
        self.outfile.close()
        self.devnull.close()
        if retcode != 0:
            raise IOError(
                "Output gzip process terminated with exit code {0}".format(
                    retcode))
    
    def __enter__(self):
        return self
    
    def __exit__(self, *exc_info):
        self.close()

class GzipReader:
    """Wrapper for a process that uses the system gzip program to decompress
    bytes.
    
    Args:
        path: The path of the input file.
    """
    def __init__(self, path):
        self.name = path
        self.process = Popen([get_program_path('gzip'), '-cd', path], stdout=PIPE)
        self.closed = False
    
    def readable(self):
        return True
    
    def writable(self):
        return False
    
    def seekable(self):
        return False
    
    def flush(self):
        pass
    
    def close(self):
        if self.closed:
            return
        self.closed = True
        retcode = self.process.poll()
        if retcode is None:
            # still running
            self.process.terminate()
        self._raise_if_error()
    
    def __iter__(self):
        for line in self.process.stdout:
            yield line
        self.process.wait()
        self._raise_if_error()
    
    def _raise_if_error(self):
        """Raise EOFError if process is not running anymore and the exit code
        is nonzero.
        """
        retcode = self.process.poll()
        if retcode is not None and retcode != 0:
            raise EOFError(
                "gzip process returned non-zero exit code {0}. Is the "
                "input file truncated or corrupt?".format(retcode))
    
    def read(self, *args):
        data = self.process.stdout.read(*args)
        if len(args) == 0 or args[0] <= 0:
            # wait for process to terminate until we check the exit code
            self.process.wait()
        self._raise_if_error()
        return data
    
    def __enter__(self):
        return self
    
    def __exit__(self, *exc_info):
        self.close()

def can_use_system_compression():
    """Whether the system gzip program is available.
    """
    return get_program_path("gzip") is not None

def get_compressor(filename):
    """Returns the python compression library for a file based on its extension.
    """
    ext = os.path.splitext(filename)[1]
    if ext in COMPRESSORS:
        return COMPRESSORS[ext]
    return None

def open_gzip_file(filename, mode, use_system=True):
    """Open a gzip file, preferring the system gzip program if `use_system`
    is True, falling back to the gzip python library.
    
    Args:
        mode: The file open mode.
        use_system: Whether to try to use the system gzip program.
    """
    if use_system:
        try:
            if 'r' in mode:
                gzfile = GzipReader(filename)
            else:
                gzfile = GzipWriter(filename)
            if 't' in mode:
                gzfile = io.TextIOWrapper(gzfile)
            return gzfile
        except:
            pass
    
    gzfile = gzip.open(filename, mode)
    if 'b' in mode:
        if 'r' in mode:
            gzfile = io.BufferedReader(gzfile)
        else:
            gzfile = io.BufferedWriter(gzfile)
    return gzfile

def open_bzip_file(filename, mode, **kwargs):
    """Open a bzip file.
    """
    if 't' in mode:
        return io.TextIOWrapper(bz2.BZ2File(filename, mode[0]))
    else:
        return bz2.BZ2File(filename, mode)

def open_lzma_file(filename, mode, **kwargs):
    """Open a LZMA (xz) file.
    """
    return lzma.open(filename, mode)

FILE_OPENERS = {
    ".gz"  : open_gzip_file,
    ".bz2" : open_bzip_file,
    ".xz"  : open_lzma_file,
}
"""Mapping of file extensions to file opener functions."""

def get_file_opener(filename):
    """Returns the file opener for a filename based on its extension.
    """
    ext = os.path.splitext(filename)[1]
    if ext in FILE_OPENERS:
        return FILE_OPENERS[ext]
    return None

PROGRAM_CACHE = {}
"""Cache for resolved program paths."""

def get_program_path(program):
    """Get the path of a program on the system. Returns from the cache if it
    was previously resolved, otherwise finds it in the system path and caches
    it.
    
    Args:
        program: The command name to resolve.
    
    Returns:
        The fully resolved path of the command.
    """
    if program in PROGRAM_CACHE:
        return PROGRAM_CACHE[program]
    
    def is_exe(fpath):
        """Returns True if `fpath` is executable.
        """
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
    
    exe_file = None
    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            exe_file = program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                break
    
    PROGRAM_CACHE[program] = exe_file
    return exe_file

def open_compressed_file(filename, mode):
    """Open a compressed file, determining the compression type based on the
    file name.
    
    Args:
        filename: The file to open.
        mode: The file open mode.
    
    Returns:
        The opened file.
    """
    ext = os.path.splitext(filename)
    opener = get_file_opener(ext)
    if not opener:
        raise ValueError("{} is not a recognized compression format")
    return opener(filename, mode)

def splitext_compressed(name):
    """Split the filename and extensions of a file that potentially has two
    extensions - one for the file type (e.g. 'fq') and one for the compression
    type (e.g. 'gz').
    
    Args:
        name: The filename.
    
    Returns:
        A tuple (name, ext1, ext2), where ext1 is the filetype extension and
        ext2 is the compression type extension, or None.
    """
    ext1 = ext2 = None
    for ext in COMPRESSORS:
        if name.endswith(ext):
            ext2 = ext
            name = name[:-len(ext)]
            break
    name, ext1 = os.path.splitext(name)
    return (name, ext1, ext2)



# coding: utf-8
import os
import shutil
from unittest import TestCase
from pytest import raises
from atropos.commands import execute_cli, get_command
from tests.utils import (
    run, files_equal, datapath, cutpath, redirect_stderr, temporary_path
)

BACK_ALIGNERS = ('adapter', 'insert')


def run_paired(
    params,
    in1,
    in2,
    expected1,
    expected2,
    aligners=('adapter',),
    callback=None,
    assert_files_equal=True,
    error_on_rc=True,
):
    if type(params) is str:
        params = params.split()
    for aligner in aligners:
        with temporary_path('tmp1-' + expected1.format(aligner=aligner)) as p1:
            with temporary_path('tmp2-' + expected2.format(aligner=aligner)) as p2:
                p = params.copy()
                p += ['--aligner', aligner, '-o', p1, '-p', p2]
                infiles = [datapath(i.format(aligner=aligner)) for i in (in1, in2)]
                for infile_args in zip(('-pe1', '-pe2'), infiles):
                    p.extend(infile_args)
                command = get_command('trim')
                result = command.execute(p)
                assert isinstance(result, tuple)
                assert len(result) == 2
                if error_on_rc:
                    err = result[1]['exception'] if result[1] and 'exception' in result[
                        1
                    ] else None
                    # from traceback import format_exception
                    if result[0] != 0:
                        if err is None:
                            raise AssertionError(
                                "Return code {} != 0".format(result[0])
                            )

                        else:
                            raise AssertionError(
                                "Return code {} != 0".format(result[0])
                            ) from err[
                                'details'
                            ][
                                1
                            ]

                if assert_files_equal:
                    assert files_equal(cutpath(expected1.format(aligner=aligner)), p1)
                    assert files_equal(cutpath(expected2.format(aligner=aligner)), p2)
                if callback:
                    callback(aligner, infiles, (p1, p2), result)


def run_interleaved(params, inpath, expected, aligners=('adapter',)):
    if type(params) is str:
        params = params.split()
    for aligner in aligners:
        with temporary_path(expected.format(aligner=aligner)) as tmp:
            p = params.copy()
            p += [
                '--aligner',
                aligner,
                '-l',
                datapath(inpath.format(aligner=aligner)),
                '-L',
                tmp,
            ]
            command = get_command('trim')
            result = command.execute(p)
            assert isinstance(result, tuple)
            assert len(result) == 2
            assert files_equal(cutpath(expected.format(aligner=aligner)), tmp)


# def run_interleaved2(params, inpath, expected1, expected2, aligners=('adapter',)):
#     assert False  # unused function
#     if type(params) is str:
#         params = params.split()
#     for aligner in aligners:
#         with temporary_path('tmp1-' + expected1.format(aligner=aligner)) as p1:
#             with temporary_path('tmp2-' + expected2.format(aligner=aligner)) as p2:
#                 p = params.copy()
#                 p += ['--aligner', aligner, '--interleaved', '-o', p1, '-p', p2]
#                 p += [datapath(inpath.format(aligner=aligner))]
#                 assert execute_cli(p) == 0
#                 assert files_equal(cutpath(expected.format(aligner=aligner)), p1)
#                 assert files_equal(cutpath(expected.format(aligner=aligner)), p2)


def test_paired_separate():
    """test separate trimming of paired-end reads"""
    run('-a TTAGACATAT', 'paired-separate.1.fastq', 'paired.1.fastq')
    run('-a CAGTGGAGTA', 'paired-separate.2.fastq', 'paired.2.fastq')


def test_paired_end_legacy():
    """--paired-output, not using -A/-B/-G"""
    # The -m 14 filters out one read, which should then also be filtered out in
    # the second output file.
    run_paired(
        '-a TTAGACATAT -m 14',
        in1='paired.1.fastq',
        in2='paired.2.fastq',
        expected1='paired.m14.1.fastq',
        expected2='paired.m14.2.fastq',
    )


def test_untrimmed_paired_output():
    with temporary_path("tmp-untrimmed.1.fastq") as untrimmed1:
        with temporary_path("tmp-untrimmed.2.fastq") as untrimmed2:

            def callback(aligner, infiles, outfiles, result):
                assert files_equal(cutpath('paired-untrimmed.1.fastq'), untrimmed1)
                assert files_equal(cutpath('paired-untrimmed.2.fastq'), untrimmed2)

            run_paired(
                [
                    '-a',
                    'TTAGACATAT',
                    '--untrimmed-output',
                    untrimmed1,
                    '--untrimmed-paired-output',
                    untrimmed2,
                ],
                in1='paired.1.fastq',
                in2='paired.2.fastq',
                expected1='paired-trimmed.1.fastq',
                expected2='paired-trimmed.2.fastq',
                callback=callback,
            )


def test_explicit_format_with_paired():
    # Use --format=fastq with input files whose extension is .txt
    with temporary_path("paired.1.txt") as txt1:
        with temporary_path("paired.2.txt") as txt2:
            shutil.copyfile(datapath("paired.1.fastq"), txt1)
            shutil.copyfile(datapath("paired.2.fastq"), txt2)
            run_paired(
                '--format=fastq -a TTAGACATAT -m 14',
                in1=txt1,
                in2=txt2,
                expected1='paired.m14.1.fastq',
                expected2='paired.m14.2.fastq',
            )


def test_no_trimming_legacy():
    # make sure that this doesn't divide by zero
    execute_cli(
        [
            '-a',
            'XXXXX',
            '-o',
            '/dev/null',
            '-p',
            '/dev/null',
            '-pe1',
            datapath('paired.1.fastq'),
            '-pe2',
            datapath('paired.2.fastq'),
        ]
    )


def test_no_trimming():
    # make sure that this doesn't divide by zero
    execute_cli(
        [
            '-a',
            'XXXXX',
            '-A',
            'XXXXX',
            '-o',
            '/dev/null',
            '-p',
            '/dev/null',
            '-pe1',
            datapath('paired.1.fastq'),
            '-pe2',
            datapath('paired.2.fastq'),
        ]
    )


def test_missing_file():
    with raises(SystemExit), redirect_stderr():
        execute_cli(
            ['-a', 'XX', '--paired-output', 'out.fastq', datapath('paired.1.fastq')]
        )


def test_first_too_short():
    with temporary_path("truncated.1.fastq") as trunc1:
        # Create a truncated file in which the last read is missing
        with open(datapath('paired.1.fastq')) as f:
            lines = f.readlines()
            lines = lines[:-4]
        with open(trunc1, 'w') as f:
            f.writelines(lines)
        with raises(SystemExit), redirect_stderr():
            execute_cli(
                '-a XX --paired-output out.fastq'.split() +
                [trunc1, datapath('paired.2.fastq')]
            )


def test_second_too_short():
    with temporary_path("truncated.2.fastq") as trunc2:
        # Create a truncated file in which the last read is missing
        with open(datapath('paired.2.fastq')) as f:
            lines = f.readlines()
            lines = lines[:-4]
        with open(trunc2, 'w') as f:
            f.writelines(lines)
        with raises(SystemExit), redirect_stderr():
            execute_cli(
                '-a XX --paired-output out.fastq'.split() +
                [datapath('paired.1.fastq'), trunc2]
            )


def test_unmatched_read_names():
    with temporary_path("swapped.1.fastq") as swapped:
        try:
            # Create a file in which reads 2 and are swapped
            with open(datapath('paired.1.fastq')) as f:
                lines = f.readlines()
                lines = lines[0:4] + lines[8:12] + lines[4:8] + lines[12:]
            with open(swapped, 'w') as f:
                f.writelines(lines)
            with redirect_stderr():
                command = get_command('trim')
                result = command.execute(
                    '-a XX -o out1.fastq -p out2.fastq'.split() +
                    ['-pe1', swapped, '-pe2', datapath('paired.2.fastq')]
                )
                assert isinstance(result, tuple)
                assert len(result) == 2
                assert result[0] != 0
        finally:
            os.remove('out1.fastq')
            os.remove('out2.fastq')


def test_legacy_minlength():
    """Ensure -m is not applied to second read in a pair in legacy mode"""
    run_paired(
        '-a XXX -m 27',
        in1='paired.1.fastq',
        in2='paired.2.fastq',
        expected1='paired-m27.1.fastq',
        expected2='paired-m27.2.fastq',
    )


def test_paired_end():
    """single-pass paired-end with -m"""
    run_paired(
        '-a TTAGACATAT -A CAGTGGAGTA -m 14',
        in1='paired.1.fastq',
        in2='paired.2.fastq',
        expected1='paired_{aligner}.1.fastq',
        expected2='paired_{aligner}.2.fastq',
        aligners=BACK_ALIGNERS,
    )


def test_paired_anchored_back_no_indels():
    run_paired(
        "-a BACKADAPTER$ -A BACKADAPTER$ -N --no-indels",
        in1='anchored-back.fasta',
        in2='anchored-back.fasta',
        expected1='anchored-back.fasta',
        expected2="anchored-back.fasta",
    )


def test_paired_end_qualtrim():
    """single-pass paired-end with -q and -m"""
    run_paired(
        '-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90',
        in1='paired.1.fastq',
        in2='paired.2.fastq',
        expected1='pairedq.1.fastq',
        expected2='pairedq.2.fastq',
        aligners=BACK_ALIGNERS,
    )


def test_paired_end_qualtrim_swapped():
    """single-pass paired-end with -q and -m, but files swapped"""
    run_paired(
        '-q 20 -a CAGTGGAGTA -A TTAGACATAT -m 14 --adapter-max-rmp 0.001',
        in1='paired.2.fastq',
        in2='paired.1.fastq',
        expected1='pairedq.2.fastq',
        expected2='pairedq.1.fastq',
        aligners=BACK_ALIGNERS,
    )


def test_paired_end_cut():
    run_paired(
        '-u 3 -u -1 -U 4 -U -2',
        in1='paired.1.fastq',
        in2='paired.2.fastq',
        expected1='pairedu.1.fastq',
        expected2='pairedu.2.fastq',
    )


def test_paired_end_A_only():
    run_paired(
        '-A CAGTGGAGTA',
        in1='paired.1.fastq',
        in2='paired.2.fastq',
        expected1='paired-onlyA.1.fastq',
        expected2='paired-onlyA.2.fastq',
    )


def test_paired_end_mask_adapter():
    """mask adapter with N (reads maintain the same length)"""
    run_paired(
        "-a CAAG -A TCGA -n 3 --mask-adapter",
        in1="back_repeat.1.fastq",
        in2="back_repeat.2.fastq",
        expected1="back_repeat.1.fastq",
        expected2="back_repeat.2.fastq",
        aligners=BACK_ALIGNERS,
    )


def test_discard_untrimmed():
    # issue #146
    # the first adapter is a sequence cut out from the first read
    run_paired(
        '-a CTCCAGCTTAGACATATC -A XXXXXXXX --discard-untrimmed',
        in1='paired.1.fastq',
        in2='paired.2.fastq',
        expected1='empty.fastq',
        expected2='empty.fastq',
    )


def test_discard_trimmed():
    run_paired(
        '-A C -O 1 --discard-trimmed',  # applies everywhere
        in1='paired.1.fastq',
        in2='paired.2.fastq',
        expected1='empty.fastq',
        expected2='empty.fastq',
    )


def test_interleaved():
    """single-pass interleaved paired-end with -q and -m"""
    run_interleaved(
        '-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90',
        inpath='interleaved.fastq',
        expected='interleaved.fastq',
        aligners=BACK_ALIGNERS,
    )


def test_interleaved_no_paired_output():
    with temporary_path("temp-paired.1.fastq") as p1:
        with temporary_path("temp-paired.2.fastq") as p2:
            params = '-a XX --interleaved'.split()
            with raises(SystemExit), redirect_stderr():
                params += ['-o', p1, '-p1', p2, 'paired.1.fastq', 'paired.2.fastq']
                execute_cli(params)




# TODO
# def test_interleaved_input_paired_output():
#     """single-pass interleaved paired-end with -q and -m, paired output"""
#     run_interleaved2('-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90',
#         inpath='interleaved.fastq', expected1='pairedq1.fastq', expected2='pairedq2.fastq'
#     )
def test_pair_filter():
    run_paired(
        '--pair-filter=both -a TTAGACATAT -A GGAGTA -m 14',
        in1='paired.1.fastq',
        in2='paired.2.fastq',
        expected1='paired-filterboth_{aligner}.1.fastq',
        expected2='paired-filterboth_{aligner}.2.fastq',
        aligners=BACK_ALIGNERS,
    )


def test_too_short_paired_output():
    with temporary_path("temp-too-short.1.fastq") as p1:
        with temporary_path("temp-too-short.2.fastq") as p2:

            def callback(aligner, infiles, outfiles, result):
                assert files_equal(cutpath('paired-too-short.1.fastq'), p1)
                assert files_equal(cutpath('paired-too-short.2.fastq'), p2)

            run_paired(
                '-a TTAGACATAT -A CAGTGGAGTA -m 14 --too-short-output '
                '{0} --too-short-paired-output {1}'.format(p1, p2),
                in1='paired.1.fastq',
                in2='paired.2.fastq',
                expected1='paired_{aligner}.1.fastq',
                expected2='paired_{aligner}.2.fastq',
                aligners=BACK_ALIGNERS,
                callback=callback,
            )


def test_too_long_output():
    with temporary_path("temp-too-long.1.fastq") as p1:
        with temporary_path("temp-too-long.2.fastq") as p2:

            def callback(aligner, infiles, outfiles, result):
                assert files_equal(
                    cutpath('paired_{aligner}.1.fastq'.format(aligner=aligner)), p1
                )
                assert files_equal(
                    cutpath('paired_{aligner}.2.fastq'.format(aligner=aligner)), p2
                )

            run_paired(
                '-a TTAGACATAT -A CAGTGGAGTA -M 14 --too-long-output '
                '{0} --too-long-paired-output {1}'.format(p1, p2),
                in1='paired.1.fastq',
                in2='paired.2.fastq',
                expected1='paired-too-short.1.fastq',
                expected2='paired-too-short.2.fastq',
                aligners=BACK_ALIGNERS,
                callback=callback,
            )


def test_too_short_output_paired_option_missing():
    with raises(SystemExit), temporary_path("temp-too-short.1.fastq") as p1:
        run_paired(
            '-a TTAGACATAT -A CAGTGGAGTA -m 14 --too-short-output ' '{0}'.format(p1),
            in1='paired.1.fastq',
            in2='paired.2.fastq',
            expected1='paired.1.fastq',
            expected2='paired.2.fastq',
            aligners=BACK_ALIGNERS,
        )


def test_custom_bisulfite_1():
    run_paired(
        '-a TTAGACATAT -A CAGTGGAGTA -m 14 -q 0 --bisulfite 2,2,1,1',
        in1='paired_bis_{aligner}.1.fastq',
        in2='paired_bis_{aligner}.2.fastq',
        expected1='paired_bis1_{aligner}.1.fastq',
        expected2='paired_bis1_{aligner}.2.fastq',
        aligners=BACK_ALIGNERS,
    )


def test_custom_bisulfite_2():
    run_paired(
        '-a TTAGACATAT -A CAGTGGAGTA -m 10 -q 0 --bisulfite 20,20,1,1;0,0,0,0',
        in1='paired_bis_{aligner}.1.fastq',
        in2='paired_bis_{aligner}.2.fastq',
        expected1='paired_bis2_{aligner}.1.fastq',
        expected2='paired_bis2_{aligner}.2.fastq',
        aligners=BACK_ALIGNERS,
    )


def test_no_insert_match():
    # with -O
    # Note: this fails if you set -e 0.3 because the higher error rate enables a
    # match of 7 bp, even though it has 2 errors. This illustrates why using
    # --adapter-max-rmp is better.
    # run_paired('-a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT -O 7 -m 25 -q 0 --trim-n',
    #     in1='insert.1.fastq', in2='insert.2.fastq',
    #     expected1='insert.1.fastq', expected2='insert.2.fastq',
    #     aligners=('insert',)
    # )
    # with --adapter-max-rmp
    run_paired(
        '-a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT -e 0.3 --adapter-max-rmp 0.001 -m 25 -q 0 --trim-n',
        in1='insert.1.fastq',
        in2='insert.2.fastq',
        expected1='insert.1.fastq',
        expected2='insert.2.fastq',
        aligners=('insert',),
    )


def test_overwrite():
    run_paired(
        '-w 10,30,10',
        in1='lowq.fastq',
        in2='highq.fastq',
        expected1='lowq.fastq',
        expected2='highq.fastq',
    )


def test_no_writer_process():

    def check_multifile(aligner, infiles, outfiles, result):
        assert os.path.basename(outfiles[0]) == 'tmp1-out.1.fastq'
        assert os.path.basename(outfiles[1]) == 'tmp2-out.2.fastq'
        tmpdir = os.path.dirname(outfiles[0])
        assert tmpdir == os.path.dirname(outfiles[1])
        # TODO: If the final worker doesn't get the chance to process any
        # batches, the last output file is never created.
        assert os.path.exists(os.path.join(tmpdir, 'tmp1-out.1.0.fastq'))
        assert os.path.exists(os.path.join(tmpdir, 'tmp1-out.1.1.fastq'))
        # assert os.path.exists(os.path.join(tmpdir, 'tmp1-out.1.2.fastq'))
        assert os.path.exists(os.path.join(tmpdir, 'tmp2-out.2.0.fastq'))
        assert os.path.exists(os.path.join(tmpdir, 'tmp2-out.2.1.fastq'))

    # assert os.path.exists(os.path.join(tmpdir, 'tmp2-out.2.2.fastq'))
    # TODO: check contents
    run_paired(
        '--threads 3 --no-writer-process --batch-size 1 -a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT',
        in1='big.1.fq',
        in2='big.2.fq',
        expected1='out.1.fastq',
        expected2='out.2.fastq',
        aligners=BACK_ALIGNERS,
        assert_files_equal=False,
        callback=check_multifile,
    )


def test_summary():

    def check_summary(aligner, infiles, outfiles, result):
        summary = result[1]
        assert summary is not None
        assert isinstance(summary, dict)
        assert summary['command'] == 'trim'
        assert summary['options']['orig_args'] == [
            '--threads',
            '2',
            '-a',
            'AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG',
            '-A',
            'AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT',
            '--aligner',
            aligner,
            '-o',
            outfiles[0],
            '-p',
            outfiles[1],
            '-pe1',
            infiles[0],
            '-pe2',
            infiles[1],
        ]
        assert summary['sample_id'] == 'big'
        assert summary['mode'] == 'parallel'
        assert summary['threads'] == 2
        # infiles are 100 125 bp PE reads
        expected_record_counts = {}
        expected_record_counts[0] = 100
        assert summary['record_counts'] == expected_record_counts
        expected_bp_counts = {}
        expected_bp_counts[0] = [12500, 12500]
        assert summary['bp_counts'] == expected_bp_counts
        assert 'timing' in summary
        assert 'wallclock' in summary['timing']
        assert summary['timing']['wallclock'] > 0
        assert 'cpu' in summary['timing']
        assert summary['timing']['cpu'] > 0

    run_paired(
        '--threads 2 -a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT',
        in1='big.1.fq',
        in2='big.2.fq',
        expected1='out.1.fastq',
        expected2='out.2.fastq',
        aligners=BACK_ALIGNERS,
        assert_files_equal=False,
        callback=check_summary,
    )


class IssueTests(TestCase):
    def test_issue68(self):
        run_paired(
            '--error-rate 0.20 --insert-match-error-rate 0.30 --minimum-length 20 --aligner insert -a AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT',
            in1='issue68.1.fq',
            in2='issue68.2.fq',
            expected1='issue68.1.fq',
            expected2='issue68.2.fq',
            aligners=['insert'],
        )



import logging
from multiprocessing import Process
import os
from atropos.commands.trim import (
    ResultHandler, WorkerResultHandler, WriterResultHandler)
from atropos.commands.multicore import (
    Control, PendingQueue, ParallelPipelineRunner, MulticoreError, 
    wait_on_process, enqueue, dequeue, kill, CONTROL_ACTIVE, CONTROL_ERROR)
from atropos.io.compression import get_compressor

class Done(MulticoreError):
    """Raised when process exits normally.
    """
    pass

class Killed(MulticoreError):
    """Raised when process exits is killed.
    """
    pass

class ParallelTrimPipelineRunner(ParallelPipelineRunner):
    """ParallelPipelineRunner for a TrimPipeline.
    """
    def __init__(
            self, command_runner, pipeline, threads, writer_manager=None):
        super().__init__(command_runner, pipeline, threads)
        self.writer_manager = writer_manager
    
    def ensure_alive(self):
        super().ensure_alive()
        if self.writer_manager and not self.writer_manager.is_active():
            raise MulticoreError("Writer process exited")
    
    def after_enqueue(self):
        # Tell the writer thread the max number of batches to expect
        if self.writer_manager:
            self.writer_manager.set_num_batches(self.num_batches)
    
    def finish(self):
        if self.writer_manager:
            # Wait for writer to complete
            self.writer_manager.wait()
    
    def terminate(self, retcode):
        super().terminate(retcode)
        if self.writer_manager:
            self.writer_manager.terminate(retcode)

class QueueResultHandler(ResultHandler):
    """ResultHandler that writes results to the output queue.
    """
    def __init__(self, queue):
        self.queue = queue
        self.message = None
        self.timeout = None
    
    def start(self, worker):
        self.message = "{} waiting to queue result {{}}".format(
            worker.name)
        self.timeout = worker.timeout
    
    def write_result(self, batch_num, result):
        enqueue(
            self.queue,
            (batch_num, result),
            wait_message=self.message,
            timeout=self.timeout)

class CompressingWorkerResultHandler(WorkerResultHandler):
    """Wraps a ResultHandler and compresses results prior to writing.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.file_compressors = None
    
    def start(self, worker):
        super().start(worker)
        self.file_compressors = {}
    
    def prepare_file(self, path, strings):
        compressor = self.get_compressor(path)
        if compressor:
            return ((path, 'wb'), compressor.compress(b''.join(
                s.encode() for s in strings)))
        else:
            return ((path, 'wt'), "".join(strings))
    
    def get_compressor(self, filename):
        """Returns the file compressor based on the file extension.
        """
        if filename not in self.file_compressors:
            self.file_compressors[filename] = get_compressor(filename)
        return self.file_compressors[filename]

class OrderPreservingWriterResultHandler(WriterResultHandler):
    """Writer thread that is less time/memory efficient, but is
    guaranteed to preserve the original order of records.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pending = None
        self.cur_batch = None
    
    def start(self, worker=None):
        super().start(worker)
        self.pending = PendingQueue()
        self.cur_batch = 1
    
    def write_result(self, batch_num, result):
        if batch_num == self.cur_batch:
            self.writers.write_result(result, self.compressed)
            self.cur_batch += 1
            self.consume_pending()
        else:
            self.pending.push(batch_num, result)
    
    def finish(self, total_batches=None):
        if total_batches is not None:
            self.consume_pending()
            if self.cur_batch != total_batches + 1:
                raise MulticoreError(
                    "OrderPreservingWriterResultHandler finishing "
                    "without having seen {} of {} batches".format(
                        total_batches + 1 - self.cur_batch,
                        total_batches))
        super().finish(total_batches=total_batches)
    
    def consume_pending(self):
        """Consume any remaining items in the queue.
        """
        while (
                (not self.pending.empty) and
                (self.cur_batch == self.pending.min_priority)):
            self.writers.write_result(
                self.pending.pop(), self.compressed)
            self.cur_batch += 1

class ResultProcess(Process):
    """Thread that accepts results from the worker threads and process
    them using a ResultHandler. Each batch is expected to be
    (batch_num, path, records), where path is the destination file and
    records is a string. Not guaranteed to preserve the original order
    of sequence records.
    
    Args:
        result_handler: A ResultHandler object.
        queue: Input queue.
        control: A shared value for communcation with the main process.
        timeout: Seconds to wait for next batch before complaining.
    """
    def __init__(self, result_handler, queue, control, timeout=60):
        super().__init__(name="Result process")
        self.result_handler = result_handler
        self.queue = queue
        self.control = control
        self.timeout = timeout
        self.seen_batches = set()
        self.num_batches = None
    
    def run(self):
        logging.getLogger().debug(
            "Writer process %s running under pid %d",
            self.name, os.getpid())
        
        def fail_callback():
            """Raises Done if the expected number of batches has been
            seen.
            """
            if (
                    self.num_batches is None and
                    self.control.check_value_positive()):
                self.num_batches = self.control.get_value()
            if (
                    self.num_batches is not None and
                    len(self.seen_batches) >= self.num_batches):
                raise Done()
        
        def timeout_callback():
            """Logs an error with the missing batches.
            """
            if self.num_batches is not None:
                missing = (
                    set(range(1, self.num_batches+1)) -
                    self.seen_batches)
                logging.getLogger().error(
                    "Result thread still missing batches %s of %d",
                    ",".join(str(i) for i in missing),
                    self.num_batches)
        
        def iter_batches():
            """Dequeue and yield batches.
            """
            while True:
                batch = dequeue(
                    self.queue,
                    wait_message="Result process waiting on result {}",
                    timeout=self.timeout,
                    fail_callback=fail_callback,
                    timeout_callback=timeout_callback)
                yield batch
        
        try:
            self.result_handler.start(self)
            
            for batch_num, result in iter_batches():
                self.seen_batches.add(batch_num)
                self.result_handler.write_result(batch_num, result)
        except Done:
            logging.getLogger().debug("Writer process exiting normally")
        except Killed:
            logging.getLogger().debug("Writer process exited early")
        except:
            logging.getLogger().error(
                "Unexpected error in writer process", exc_info=True)
            self.control.set_value(CONTROL_ERROR)
        finally:
            num_batches = self.control.get_value(lock=True)
            self.result_handler.finish(
                num_batches if num_batches > 0 else None)

class WriterManager(object):
    """Manager for a writer process and control variable.
    """
    def __init__(
            self, writers, compression, preserve_order, result_queue,
            timeout):
        # result handler
        if preserve_order:
            writer_result_handler = OrderPreservingWriterResultHandler(
                writers, compressed=compression == "worker")
        else:
            writer_result_handler = WriterResultHandler(
                writers, compressed=compression == "worker")
        
        self.timeout = timeout
        # Shared variable for communicating with writer thread
        self.writer_control = Control(CONTROL_ACTIVE)
        # writer process
        self.writer_process = ResultProcess(
            writer_result_handler, result_queue, self.writer_control,
            timeout)
        self.writer_process.start()
    
    def is_active(self):
        """Returns True if the writer process is alive and the control
        value is CONTROL_ACTIVE.
        """
        return (
            self.writer_process.is_alive() and
            self.writer_control.check_value(CONTROL_ACTIVE))
    
    def set_num_batches(self, num_batches):
        """Set the number of batches to the control variable.
        """
        self.writer_control.set_value(num_batches)
    
    def wait(self):
        """Wait for the writer process to terminate.
        """
        wait_on_process(self.writer_process, self.timeout)
    
    def terminate(self, retcode):
        """Force the writer process to terminate.
        """
        kill(self.writer_process, retcode, self.timeout)


# Generate base histograms of the first and last 20 bp of read1 and read2

import argparse
from common import *
import csv
from xphyle import xopen
import tqdm

nuc = ('A','C','G','T','N')

def make_hists(fq1, fq2, n_bases=20):
    read1_hist = [
        dict((base, [0] * n_bases) for base in nuc)
        for i in range(2)
    ]
    read2_hist = [
        dict((base, [0] * n_bases) for base in nuc)
        for i in range(2)
    ]
    
    def add_to_hist(r, hist):
        for i, b in enumerate_range(r[2].upper(), 0, n_bases):
            hist[0][b][i] += 1
        for i, b in enumerate_range(reversed(r[2].upper()), 0, n_bases):
            hist[1][b][i] += 1
    
    for r1, r2 in tqdm.tqdm(zip(fq_iterator(fq1), fq_iterator(fq2))):
        add_to_hist(r1, read1_hist)
        add_to_hist(r2, read2_hist)
    
    return (read1_hist, read2_hist)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-1", "--fastq1")
    parser.add_argument("-2", "--fastq2")
    parser.add_argument("-o", "--output", default="-")
    args = parser.parse_args()
    
    with xopen(args.fastq1) as fq1, xopen(args.fastq2) as fq2:
        hists = make_hists(fq1, fq2)
    
    with xopen(args.output, 'w') as o:
        w = csv.writer(o, delimiter="\t")
        w.writerow(('read', 'side', 'pos', 'base', 'count'))
        for i, h in enumerate(hists, 1):
            for j in range(2):
                for b in nuc:
                    for k, count in enumerate(h[j][b], 1):
                        w.writerow((i, j, k, b, count))

if __name__ == "__main__":
    main()



"""Detect adapter sequences directly from reads based on kmer frequency.
"""
from collections import defaultdict
import logging
import math
import re
from atropos.align import Aligner, SEMIGLOBAL
from atropos.commands.base import (
    BaseCommandRunner, Pipeline, SingleEndPipelineMixin, PairedEndPipelineMixin)
from atropos.util import (
    reverse_complement, sequence_complexity, run_interruptible)

# TODO: Test whether using rc=True in parse_known_contaminants is as fast
# and as accurate as testing both the forward and reverse complement
# read sequence against only the forward-orientation known contaminants.
# Also, offer an option of whether to test the reverse complement, with
# the default being false.

# TODO: I'm not sure that the use of tags in the KhmerDetector is the correct
# strategy, mostly because I don't fully understand what tags are (they are not
# documented and only mentioned briefly in the PNAS paper). An alternate
# strategy is to use CountTable and, upon each kmer addition, test whether its
# count exceeds some threshold (perhaps by maintaining a running distribution of
# all kmer counts). These high-abundance kmers could then be profiled using a
# second pass. Another strategy is to compute summary metrics for reads based on
# kmer abundance, as per
# http://khmer-recipes.readthedocs.io/en/latest/001-extract-reads-by-coverage/index.html
# Then take reads in the furthest-right peak and extract common sequences from
# those.

# TODO: Check out AdapterRemoval2's detection strategy.

# pymer https://github.com/kdmurray91/pymer/tree/master/pymer
# ntCard https://github.com/bcgsc/ntCard
# https://github.com/TGAC/KAT
# https://arxiv.org/pdf/1707.01743.pdf
# https://github.com/splatlab/squeakr
# bounter is a general-purpose counter

# TODO: In KnownContaminantDetector, accept template sequences with wildcards
# to match against.

# TODO: Re-download sequencing_adapters.fa if it has been updated since last
# download.

# TODO: parallelize adapter detection for multiple input files.


class CommandRunner(BaseCommandRunner):
    """

    """
    name = 'detect'

    def __call__(self):
        kmer_size = self.kmer_size or 12
        n_reads = self.max_reads
        overrep_cutoff = 100
        include = self.include_contaminants or "all"
        known_contaminants = None
        if include != 'unknown':
            known_contaminants = self.load_known_adapters()

        detector = self.detector
        if not detector:
            if known_contaminants and include == 'known':
                detector = 'known'
            elif n_reads <= 50000:
                detector = 'heuristic'
            else:
                detector = 'khmer'

        detector_args = dict(known_contaminants=known_contaminants)

        if detector == 'known':
            logging.getLogger().debug(
                "Detecting contaminants using the known-only algorithm")
            detector_class = KnownContaminantDetector
            detector_args['min_kmer_match_frac'] = self.min_kmer_match_frac
        elif detector == 'heuristic':
            logging.getLogger().debug(
                "Detecting contaminants using the heuristic algorithm")
            detector_class = HeuristicDetector
            detector_args['min_frequency'] = self.min_frequency
            detector_args['min_contaminant_match_frac'] = \
                self.min_contaminant_match_frac
        elif detector == 'khmer':
            logging.getLogger().debug(
                "Detecting contaminants using the kmer-based algorithm")
            detector_class = KhmerDetector
        else:
            raise ValueError("Invalid value for 'detector': {}".format(detector))

        summary_args = dict(
            kmer_size=kmer_size, n_reads=n_reads,
            overrep_cutoff=overrep_cutoff, include=include,
            past_end_bases=self.past_end_bases)
        detector_args.update(summary_args)

        if self.paired:
            detector = PairedDetector(detector_class, **detector_args)
        else:
            detector = detector_class(**detector_args)

        self.summary['detect'] = summary_args
        if known_contaminants:
            self.summary['detect']['known_contaminants'] = \
                known_contaminants.summarize()

        logging.getLogger().info(
            "Detecting adapters and other potential contaminant "
            "sequences based on %d-mers in %d reads", kmer_size, n_reads)

        # currently only single-threaded operation is supproted
        self.summary.update(mode='serial', threads=1)
        return run_interruptible(detector, self, raise_on_error=True)


class Match(object):
    """A contaminant match.

    Args:
        seq_or_contam: The matched sequence.
        count: The number of matches.
        names: The name(s) of matching contaminant(s).
        match_frac: The fraction of contaminant kmers that match.
        match_frac2: The fraction of sequence kmers that match.
        reads: The number of reads with the contaminant.
    """
    def __init__(
            self, seq_or_contam, count=0, names=None, match_frac=None,
            match_frac2=None, abundance=None, reads=None):
        if isinstance(seq_or_contam, ContaminantMatcher):
            self.seq = seq_or_contam.seq
            self.count = int(seq_or_contam.matches)
            self.names = tuple(seq_or_contam.names)
            self.known_seqs = [seq_or_contam.seq]
        else:
            self.seq = seq_or_contam
            self.count = count
            self.names = tuple(names) if names else None
            self.known_seqs = None
        self.match_frac = match_frac
        self.match_frac2 = match_frac2
        self.abundance = abundance
        self.longest_match = None
        if reads:
            self.set_longest_match(reads)

    def __len__(self):
        return len(self.seq)

    def __repr__(self):
        if self.is_known:
            return "{} => {} ({}))".format(
                self.seq, self.names, self.known_seqs)
        else:
            return self.seq

    @property
    def seq_complexity(self):
        """The complexity of the sequence (0=homopolymer, 2=random).
        """
        return sequence_complexity(self.seq)

    @property
    def count_is_frequency(self):
        """Whether `self.count` represents a frequency.
        """
        return isinstance(self.count, float)

    def set_contaminant(self, contam, match_frac, match_frac2=None):
        """Set the known contaminant from a :class:`ContaminantMatcher`.
        """
        self.set_known(contam.names, [contam.seq], match_frac, match_frac2)

    def set_known(self, names, seqs, match_frac, match_frac2=None):
        """Set the known contaminant.
        """
        self.names = tuple(names) if names else None
        self.known_seqs = seqs
        self.match_frac = match_frac
        self.match_frac2 = match_frac2

    @property
    def is_known(self):
        """Whether the matching sequence is a known contaminant.
        """
        return self.known_seqs is not None

    def set_longest_match(self, sequences):
        """Set the longest matching sequence from a set of sequences.
        """
        for seq in sequences:
            idx = seq.index(self.seq)
            seqlen = len(self.seq) - idx
            if self.longest_match is None or self.longest_match[1] < seqlen:
                self.longest_match = (seq[idx:], seqlen)

    def estimate_abundance(self, read_sequences):
        """Determine whether this match's sequence is within 'seq' by simple
        exact string comparison.
        """
        self.abundance = sum(
            1 for read_seq in read_sequences
            if self.seq in read_seq)

    def summarize(self):
        summary = dict(
            longest_kmer=self.seq,
            kmer_freq=self.count,
            kmer_freq_type="frequency" if self.count_is_frequency else "count",
            abundance=self.abundance,
            is_known=self.is_known,
            known_to_contaminant_match_frac=None,
            contaminant_to_known_match_frac=None,
            longest_match=None,
            known_names=None,
            known_seqs=None)
        if self.longest_match:
            summary.update(longest_match=self.longest_match[0])
        if self.is_known:
            summary.update(
                known_to_contaminant_match_frac=self.match_frac,
                contaminant_to_known_match_frac=self.match_frac2,
                known_names=self.names,
                known_seqs=self.known_seqs)
        return summary


class ContaminantMatcher(object):
    """Matches a known contaminant against other sequences.

    Args:
        seq: The contaminant sequence.
        names: Sequence of names for the contaminant.
        kmer_size: kmer size.
    """
    def __init__(self, seq, names, kmer_size):
        self.seq = seq
        self.names = names
        self.kmers = set(
            seq[i:(i+kmer_size)]
            for i in range(len(seq) - kmer_size + 1))
        self.n_kmers = len(self.kmers)
        self.kmer_size = kmer_size
        self.matches = 0

    def match(self, seq, seqrc):
        """Returns (num_matches, num_contam_kmers, num_seq_kmers).

        Args:
            seq: The sequence to match.
            seqrc: The reverse complement of `seq`.

        Returns:
            Tuple (f1, f2, seq), where f1 is the fraction of contaminant kmers
            that match, f2 is the fraction of sequence kmers that match, and
            seq is the best matching sequence (either `seq` or `seqrc`).
        """
        fw_kmers = set(
            seq[i:(i+self.kmer_size)]
            for i in range(len(seq) - self.kmer_size + 1))
        fw_matches = float(len(self.kmers & fw_kmers))

        rv_kmers = set(
            seqrc[i:(i+self.kmer_size)]
            for i in range(len(seqrc) - self.kmer_size + 1))
        rv_matches = float(len(self.kmers & rv_kmers))

        if fw_matches >= rv_matches:
            n_matches = fw_matches
            kmers = fw_kmers
            compare_seq = seq
        else:
            n_matches = rv_matches
            kmers = rv_kmers
            compare_seq = seqrc

        self.matches += n_matches
        match_frac1 = match_frac2 = 0
        if self.n_kmers > 0:
            match_frac1 = n_matches / self.n_kmers
        if len(kmers) > 0:
            match_frac2 = n_matches / len(kmers)
        return match_frac1, match_frac2, compare_seq


def create_contaminant_matchers(contaminants, kmer_size):
    """Create :class:`ContaminantMatcher`s from sequences.

    Args:
        contaminants: A dict of {seq:names}.
        kmer_size: The kmer size.

    Returns:
        A list of :class:`ContaminantMatcher`s.
    """
    return [
        ContaminantMatcher(seq, names, kmer_size)
        for seq, names in contaminants.iter_sequences()
    ]


class Detector(SingleEndPipelineMixin, Pipeline):
    """Base class for contaminant detectors.

    Args:
        kmer_size: Size of kmers to match.
        n_reads: Number of reads to sample.
        overrep_cutoff: Degree of overrepresentation required for a kmer to be
            considered as a contaminant.
        known_contaminants: :class:`ContaminantMatcher`s to match against.
        past_end_bases: On Illumina, long runs of A (and sometimes other bases)
            can signify that the sequencer has read past the end of a fragment
            that is shorter than the read length + adapter length. Those
            bases will be removed from any sequencers before looking for
            matching contaminants.
    """
    def __init__(
            self, kmer_size=12, n_reads=10000, overrep_cutoff=100,
            include='all', known_contaminants=None, past_end_bases=('A',)):
        super().__init__()
        self.kmer_size = kmer_size
        self.n_reads = n_reads
        self.overrep_cutoff = overrep_cutoff
        self.include = include
        self.known_contaminants = known_contaminants
        self._read_length = None
        self._read_sequences = set()
        self._matches = None
        self._past_end_regexp = None
        if past_end_bases:
            if len(past_end_bases[0]) > 1:
                self._past_end_regexp = re.compile(past_end_bases[0])
            else:
                self._past_end_regexp = re.compile('|'.join(
                    base + '{8,}.*|' + base + '{2,}$'
                    for base in past_end_bases))

    @property
    def min_report_freq(self):
        """The minimum contaminant frequency required for a contaminant to be
        reported.
        """
        raise NotImplementedError()

    def set_read_length(self, record):
        assert self._read_length is None
        self._read_length = len(record.sequence)

    def handle_records(self, context, records):
        if context['size'] == 0:
            return
        if self._read_length is None:
            self.set_read_length(records[0])
        super().handle_records(context, records)

    def handle_reads(self, context, read1, read2=None):
        seq = self._filter_seq(read1.sequence)
        if seq:
            self._read_sequences.add(seq)

    def _filter_seq(self, seq):
        if sequence_complexity(seq) <= 1.0:
            return None
        if self._past_end_regexp:
            match = self._past_end_regexp.search(seq)
            if match:
                seq = seq[:match.start()]
        if len(seq) < self.kmer_size:
            return None
        return seq

    def matches(self, **kwargs):
        """Returns the current set of matches.
        """
        if self._matches is None or len(kwargs) > 0:
            self._filter_and_sort(**kwargs)
        return self._matches

    def _filter_and_sort(
            self, min_len=None, min_complexity=1.1, min_match_frac=0.1,
            limit=20):
        """Identify, filter, and sort contaminants.

        Args:
            min_len: Minimum contaminant length.
            min_complexity: Minimum sequence complexity.
            min_match_frac: Minimum fraction of matching kmers.
            limit: Maximum number of contaminants to return.
        """
        if min_len is None:
            min_len = self.kmer_size

        matches = self._get_contaminants()

        for match in matches:
            match.estimate_abundance(self._read_sequences)

        def _filter(_match):
            if _match.count < self.min_report_freq:
                logging.getLogger().debug(
                    "Filtering {} because frequency {} < {}".format(
                        _match.seq, _match.count, self.min_report_freq))
                return False
            if min_len and len(_match) < min_len:
                logging.getLogger().debug(
                    "Filtering {} because it's too short (< {} bp)".format(
                        _match.seq, min_len))
                print('too short')
                return False
            if min_complexity and _match.seq_complexity < min_complexity:
                logging.getLogger().debug(
                    "Filtering {} because its complexity {} < {}".format(
                        _match.seq, _match.seq_complexity, min_complexity))
                return False
            if self.include == 'known' and not _match.is_known:
                logging.getLogger().debug(
                    "Filtering {} because it's not known".format(
                        _match.seq))
                return False
            elif self.include == 'unknown' and _match.is_known:
                logging.getLogger().debug(
                    "Filtering {} because it's known".format(
                        _match.seq))
                return False
            if (
                    min_match_frac and _match.is_known and
                    _match.match_frac < min_match_frac):
                logging.getLogger().debug(
                    "Filtering {} because its match_frac {} < {}".format(
                        _match.seq, _match.match_frac, min_match_frac))
                return False
            return True

        matches = list(filter(_filter, matches))
        matches.sort(key=lambda x: len(x) * math.log(x.count), reverse=True)

        if limit is not None:
            matches = matches[:limit]

        self._matches = matches

    def _get_contaminants(self):
        """Implemention of contaminant matching.

        Returns:
            A list of :class:`Match`es.
        """
        raise NotImplementedError()

    def finish(self, summary, **kwargs):
        super().finish(summary)
        summary['detect']['matches'] = ([
            match.summarize()
            for match in self.matches(**kwargs)
        ],)


class PairedDetector(PairedEndPipelineMixin, Pipeline):
    """Detector for paired-end reads.
    """
    def __init__(self, detector_class, **kwargs):
        super().__init__()
        self.read1_detector = detector_class(**kwargs)
        self.read2_detector = detector_class(**kwargs)
        self._read_length_set = False

    def handle_records(self, context, records):
        if context['size'] == 0:
            return
        if not self._read_length_set:
            read1, read2 = records[0]
            self.read1_detector.set_read_length(read1)
            self.read2_detector.set_read_length(read2)
            self._read_length_set = True
        super().handle_records(context, records)

    def handle_reads(self, context, read1, read2):
        self.read1_detector.handle_reads(context, read1)
        self.read2_detector.handle_reads(context, read2)

    def finish(self, summary, **kwargs):
        super().finish(summary)
        summary['detect']['matches'] = ([
            match.summarize()
            for match in self.read1_detector.matches(**kwargs)
        ], [
            match.summarize()
            for match in self.read2_detector.matches(**kwargs)
        ])


class KnownContaminantDetector(Detector):
    """Test known contaminants against reads. This has linear complexity and is
    more specific than the khmer matcher, but less specific than the heuristic
    matcher. It's also less sensitive since it does not try to detect unknown
    contaminants.

    Args:
        known_contaminants: List of :class:`ContaminantMatcher`s.
        min_kmer_match_frac: Minimum fraction of matching kmers required.
        kwargs: Additional arguments to pass to the :class:`Detector`
            constructor.
    """
    def __init__(self, known_contaminants, min_kmer_match_frac=0.5, **kwargs):
        super().__init__(known_contaminants=known_contaminants, **kwargs)
        self.min_kmer_match_frac = min_kmer_match_frac
        self._min_k = min(len(s) for s in known_contaminants.sequences)

    @property
    def min_report_freq(self):
        return 0.1

    def _filter_seq(self, seq):
        seq = super()._filter_seq(seq)
        if seq and len(seq) >= self._min_k:
            return seq
        return None

    def _get_contaminants(self):
        contaminant_matchers = create_contaminant_matchers(
            self.known_contaminants, self.kmer_size)
        counts = defaultdict(int)
        max_match_fracs = defaultdict(int)

        for seq in self._read_sequences:
            seqrc = reverse_complement(seq)
            for contam in contaminant_matchers:
                match = contam.match(seq, seqrc)
                if match[0] > self.min_kmer_match_frac:
                    counts[contam] += 1
                    if match[0] > max_match_fracs[contam]:
                        max_match_fracs[contam] = match[0]

        min_count = math.ceil(
            self.n_reads * (self._read_length - self._min_k + 1) *
            self.overrep_cutoff / float(4**self._min_k))

        return [
            Match(
                c[0], match_frac=max_match_fracs[c[0]],
                abundance=float(c[1]) / self.n_reads)
            for c in filter(
                lambda x: x[1] >= min_count,
                counts.items()
            )
        ]


class HeuristicDetector(Detector):
    """Use a heuristic iterative algorithm to arrive at likely contaminants.
    This is the most accurate algorithm overall, but it has quadratic complexity
    and becomes too slow/memory-intenstive when n_reads > 50k.
    """
    def __init__(
            self, min_frequency=0.001, min_contaminant_match_frac=0.9,
            **kwargs):
        super(HeuristicDetector, self).__init__(**kwargs)
        self.min_frequency = min_frequency
        self.min_contaminant_match_frac = min_contaminant_match_frac

    @property
    def min_report_freq(self):
        return 0.1 * self.n_reads

    def _get_contaminants(self):
        def _min_count(_kmer_size):
            return math.ceil(self.n_reads * max(
                self.min_frequency,
                (self._read_length - _kmer_size + 1) * self.overrep_cutoff /
                float(4 ** _kmer_size)))

        kmer_size = self.kmer_size
        kmers = defaultdict(lambda: [0, set()])

        for seq in self._read_sequences:
            for i in range(len(seq) - kmer_size + 1):
                kmer = seq[i:(i+kmer_size)]
                kmers[kmer][0] += 1
                kmers[kmer][1].add(seq)

        prev = None
        cur = {}
        results = {}
        result_seqs = defaultdict(set)
        min_count = _min_count(kmer_size)

        # Identify candidate kmers for increasing values of k
        while True:
            all_seqs = set()
            for kmer, (count, seqs) in kmers.items():
                if count > min_count:
                    cur[kmer] = (count, seqs)
                    all_seqs.update(seqs)

            if len(all_seqs) == 0:
                break

            if prev:
                for kmer, (count, seqs) in prev.items():
                    if (
                            not any(seq in cur for seq in seqs) and
                            sequence_complexity(kmer) > 1.0):
                        results[kmer] = count
                        result_seqs[kmer].update(seqs)

            kmer_size += 1
            kmers = defaultdict(lambda: [0, set()])
            for seq in all_seqs:
                for i in range(len(seq) - kmer_size + 1):
                    kmer = seq[i:(i+kmer_size)]
                    kmers[kmer][0] += 1
                    kmers[kmer][1].add(seq)

            min_count = _min_count(kmer_size)
            prev = cur
            cur = {}

        results = list(results.items())

        # Now merge overlapping sequences by length and frequency to eliminate
        # redundancy in the set of candidate kmers.
        results.sort(key=lambda r: len(r[0]) * math.log(r[1]), reverse=True)
        merged = []
        unmerged = []
        while len(results) > 1:
            seq1, count1 = results[0]
            for j in range(1, len(results)):
                seq2, count2 = results[j]
                if len(seq1) >= len(seq2) and seq2 in seq1:
                    count1 += count2
                elif seq1 in seq2:
                    # if they are close in count, keep the longer sequence
                    if count1 < (2 * count2):
                        seq1 = seq2
                    count1 += count2
                else:
                    unmerged.append(results[j])
            merged.append([seq1, count1])
            results = unmerged
            unmerged = []
        results = merged + results

        if len(results) == 0:
            return []

        # TODO: For each retained match, pull out the longest sequence that
        # matches to have a better shot of identifying long adapters that
        # appear in full very infrequently

        # Re-sort by frequency
        results.sort(key=lambda r: r[1], reverse=True)
        # Keep anything that's within 50% of the top hit
        # TODO: make this user-configurable?
        min_count = int(results[0][1] * 0.5)
        results = (x for x in results if x[1] >= min_count)
        # Convert to matches
        matches = [
            Match(x[0], count=x[1], reads=result_seqs[x[0]])
            for x in results]

        if self.known_contaminants:
            # Match to known sequences
            contaminants = create_contaminant_matchers(
                self.known_contaminants, self.kmer_size)
            known = {}
            unknown = []

            def find_best_match(_seq, _best_matches, _best_match_frac):
                """Find best contaminant matches to `seq`.
                """
                seqrc = reverse_complement(_seq)

                for _contam in contaminants:
                    match_frac1, match_frac2, compare_seq = _contam.match(
                        _seq, seqrc)
                    if match_frac1 < _best_match_frac[0]:
                        continue
                    if (
                            _contam.seq in compare_seq or
                            align(
                                compare_seq, _contam.seq,
                                self.min_contaminant_match_frac)):
                        if (
                            match_frac1 > _best_match_frac[0] or (
                                match_frac1 == _best_match_frac[0] and
                                match_frac2 > _best_match_frac[1]
                            )
                        ):
                            _best_matches = {}
                            _best_match_frac = (match_frac1, match_frac2)
                        _best_matches[_contam] = (
                            match, (match_frac1, match_frac2))

                return _best_matches, _best_match_frac

            for match in matches:
                best_matches, best_match_frac = find_best_match(
                    match.seq, {}, (self.min_contaminant_match_frac, 0))

                if match.longest_match:
                    best_matches, best_match_frac = find_best_match(
                        match.longest_match[0], best_matches, best_match_frac)

                if best_matches:
                    for contam, _match in best_matches.items():
                        if contam not in known or _match[1] > known[contam][1]:
                            known[contam] = _match
                else:
                    unknown.append(match)

            # resolve many-many relationships

            new_matches = defaultdict(lambda: [])
            for contam, (match, match_frac) in known.items():
                new_matches[match].append((contam, match_frac))

            known = []
            for match, contams in new_matches.items():
                if len(contams) == 1:
                    contam, match_frac = contams[0]
                    match.set_contaminant(contam, *match_frac)
                else:
                    contams.sort(key=lambda x: x[1], reverse=True)
                    contam, match_frac = contams[0]
                    equiv = [
                        other_contam
                        for other_contam in contams[1:]
                        if other_contam[1] == match_frac]
                    if len(equiv) == 0:
                        match.set_contaminant(contam, *match_frac)
                    else:
                        names = set(contam.names)
                        seqs = {(contam.seq,)}
                        for other_contam in equiv:
                            names.update(other_contam[0].names)
                            seqs.add(other_contam[0].seq)
                        match.set_known(list(names), list(seqs), *match_frac)
                known.append(match)

            matches = known + unknown

        return matches


class KhmerDetector(Detector):
    """Identify contaminants based on kmer frequency using a fast kmer counting
    approach (as implemented in the khmer library). This approach is fast but
    not as accurate as the other two.
    """
    @property
    def min_report_freq(self):
        return 0.0001

    def _get_contaminants(self):
        from khmer import Countgraph, khmer_args
        # assuming all sequences are same length
        n_win = self._read_length - self.kmer_size + 1
        tablesize = self.n_reads * n_win
        countgraph = Countgraph(
            self.kmer_size, tablesize, khmer_args.DEFAULT_N_TABLES)
        countgraph.set_use_bigcount(True)

        for seq in self._read_sequences:
            countgraph.consume_and_tag(seq)

        n_expected = math.ceil(tablesize / float(4**self.kmer_size))
        min_count = n_expected * self.overrep_cutoff
        if min_count >= 2**16:
            raise ValueError(
                "The minimum count for an over-represented k-kmer {} is "
                "greater than the max khmer count (2^16)".format(min_count))

        candidates = {}

        for tag in countgraph.get_tagset():
            count = countgraph.get(tag)
            if count >= min_count:
                candidates[tag] = count

        if self.known_contaminants:
            matches = []
            seen = set()

            def match(_kmer):
                """Returns the frequency of `kmer` in `candidates`.
                """
                freq = candidates.get(_kmer, 0)
                if freq > 0:
                    seen.add(_kmer)
                return freq

            for seq, names in self.known_contaminants.iter_sequences():
                seqlen = len(seq)
                if seqlen < self.kmer_size:
                    print("Cannot check {}; sequence is shorter than {}".format(
                        list(names)[0], self.kmer_size))
                    continue

                n_kmers = seqlen - self.kmer_size + 1
                num_matches = 0
                match_counts = []
                for idx in range(n_kmers):
                    kmer = seq[idx:(idx + self.kmer_size)]
                    kmer_count = max(
                        match(kmer),
                        match(reverse_complement(kmer))
                    )
                    if kmer_count > 0:
                        num_matches += 1
                        match_counts.append(kmer_count)

                if num_matches > 0:
                    # not sure what the correct metric is to use here
                    overall_count = sum(match_counts) / float(n_kmers)
                    matches.append(Match(
                        seq, count=overall_count / float(tablesize),
                        names=names, match_frac=float(num_matches) / n_kmers))

            # Add remaining tags
            for tag in set(candidates.keys()) - seen:
                matches.append(Match(
                    tag, count=candidates[tag] / float(tablesize)))

        else:
            matches = [
                Match(tag, count=count / float(tablesize))
                for tag, count in candidates.items()]

        return matches


def align(seq1, seq2, min_overlap_frac=0.9):
    """Align two sequences.

    Args:
        seq1: The second sequence to align.
        seq2: The first sequence to align.
        min_overlap_frac: Minimum fraction of overlapping bases required for a
            match.

    Returns:
        The matching portion of the sequence.
    """
    aligner = Aligner(
        seq1, 0.0,
        SEMIGLOBAL,
        False, False)
    aligner.min_overlap = math.ceil(
        min(len(seq1), len(seq2)) * min_overlap_frac)
    aligner.indel_cost = 100000
    match = aligner.locate(seq2)
    if match:
        return seq1[match[0]:match[1]]
    else:
        return None



import gzip
import sys

DEFAULT_ADAPTERS = [
    "AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG",
    "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT"
]

class fileopen(object):
    def __init__(self, path, mode='wt'):
        self.close = False
        if path == '-':
            if any(m in mode for m in ('w', 'a')):
                self.fh = sys.stdout
            else:
                self.fh = sys.stdin
        elif path.endswith('.gz'):
            self.fh = gzip.open(path, mode)
            self.close = True
        else:
            self.fh = open(path, mode)
            self.close = True
    
    def __enter__(self):
        return self.fh
    
    def __exit__(self, type, value, traceback):
        if self.close:
            self.fh.close()

MAGNITUDE = dict(
    k=1000,
    m=1000000,
    g=1000000000
)

def parse_size(sizestr):
    if not sizestr:
        return -1
    mag = MAGNITUDE.get(sizestr.lower()[-1], None)
    if mag:
        return int(sizestr[0:-1]) * mag
    else:
        return int(sizestr)

def parse_profile(profile, threads=None, dataset=None):
    if profile == 'untrimmed':
        return ('untrimmed', 'untrimmed', threads, dataset, 0)
    
    profile = profile.split("_")
    
    if profile[0] == 'atropos':
        prog, threads, dataset, qcut, aligner, writer = profile
        prog = "{} ({})".format(profile[0], aligner)
        prog2 = "{} ({} + {})".format(profile[0], aligner, writer)
    else:
        prog, threads, dataset, qcut = profile
        prog2 = prog
    # If this is a simulated dataset, format the error rate
    try:
        float(dataset)
        dataset = '0.' + dataset
    except:
        pass
    qcut = qcut[1:]
    return (prog, prog2, threads, dataset, qcut)

def fq_iterator(i, pair, pair_in_name=False):
    for read in zip(*[i] * 4):
        name = read[0].rstrip()[1:]
        if pair_in_name:
            assert pair == int(name[-1])
            name = name[:-2]
        yield (name, pair, read[1].rstrip(), read[3].rstrip())

def aln_iterator(i):
    for line in i:
        if line[0] in ('@','#'):
            continue
        assert line[0] == '>'
        chrm, name, pos, strand = line[1:].rstrip().split('\t')
        if name.endswith('-1'):
            name = name[:-2]
        mate = name[-1]
        name = name[:-2]
        ref = next(i).rstrip()
        actual = next(i).rstrip()
        yield (name, mate, chrm, pos, strand, ref, actual)

def find_best_alignment(ref, query, side, min_match=1, cache=None, start=0, end=0.4, inc=0.01):
    from atropos import align
    best_match = None
    best_alternate = None
    
    for err in seq(start, end, inc):
        if cache is not None and err in cache:
            aligner = cache[err]
        else:
            aligner = make_aligner(ref, err, side)
            if cache is not None:
                cache[err] = aligner
        match = aligner.locate(query)
        if match is not None:
            ref_match = match[1] - match[0]
            if ref_match >= min_match:
                if match[5] == 0:
                    return match
                elif best_match is None or match[5] < best_match[5]:
                    best_match = match
            elif (best_alternate is None or ref_match > (best_alternate[1] - best_alternate[0]) or
                        (ref_match == (best_alternate[1] - best_alternate[0]) and match[5] < best_alternate[5])):
                best_alternate = match
    
    if best_match is not None:
        return best_match
    else:
        return best_alternate

def seq(start, end, inc):
    i = start
    while i <= end:
        yield i
        i += inc

def enumerate_range(collection, start, end):
    'Generates an indexed series:  (0,coll[0]), (1,coll[1]) ...'
    i = start
    it = iter(collection)
    while i < end:
        yield (i, next(it))
        i += 1

class BAMReader(object):
    """Reads read pairs from a name-sorted bam file."""
    def __init__(self, bam_file):
        import pysam
        self.bam = iter(pysam.AlignmentFile(bam_file, "rb"))
        self.cached = None
        self.finished = False
    
    def close(self):
        self.bam.close()
    
    def peek(self):
        if not self.cached:
            if self.finished:
                raise Exception("Iterator is exhausted")
            else:
                self.cached = next(self.bam)
        return self.cached
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        self.close()
            
    def __iter__(self):
        return self
    
    def __next__(self):
        r1 = []
        r2 = []
        def add_read(read):
            if read.is_read1:
                r1.append(read)
            else:
                r2.append(read)
        
        if self.finished:
            raise StopIteration()

        read = self.cached or next(self.bam)
        name = read.query_name
        add_read(read)
        try:
            peek = next(self.bam)
            while peek.query_name == name:
                add_read(peek)
                peek = next(self.bam)
        except StopIteration:
            self.finished = True
            peek = None
        self.cached = peek
        
        return (name, r1, r2)



"""
Atropos version {}

usage: atropos [--config <config file>] <command> [options]

commands
--------
{}

optional arguments:
  -h, --help                show this help message and exit
  --config <config file>    provide options in a config file

Use "atropos <command> --help" to see all options for a specific command.
See http://atropos.readthedocs.org/ for full documentation.

Atropos is a fork of Cutadapt 1.10 (
https://github.com/marcelm/cutadapt/tree/2f3cc0717aa9ff1e0326ea6bcb36b712950d4999)
by John Didion, et al., "Atropos: sensitive, specific, and speedy trimming of
NGS reads, submitted.

Cutadapt (https://github.com/marcelm/cutadapt) was developed by Marcel Martin,
"Cutadapt Removes Adapter Sequences From High-Throughput Sequencing Reads,"
EMBnet Journal, 2011, 17(1):10-12.
"""
from importlib import import_module
import logging
import os
from pkgutil import walk_packages
import re
import textwrap
from atropos import __version__

class Command(object):
    """Contains information about a command package.
    
    A command package consists of at least 3 modules:
    1. The top level module (__init__.py), which contains the logic to run the
    command. It must have a top-level CommandRunner class that extends
    atropos.commands.base.BaseCommandRunner.
    2. The cli module (cli.py), which configures the ArgumentParser and
    validates command line options. It must have a top-level CommandParser
    class that extends atropos.commands.base.BaseCommandParser.
    3. The reports module (reports.py), which generates reports from the
    command summary. It must have a top-level ReportGenerator class that
    extends atropos.commands.reports.BaseReportGenerator.
    
    Args:
        name: The command name. Must match the name of a submodule of
            atropos.commands, or all of the other arguments must be
            specified.
        module, cli_module, report_module: The absolute names of the three
            modules described above. If None, they are determined automatically
            from the command name.
    """
    def __init__(self, name, module=None, cli_module=None, report_module=None):
        self.name = name
        self.package = module or 'atropos.commands.{}'.format(name)
        self.cli_module = cli_module or '{}.cli'.format(self.package)
        self.report_module = report_module or '{}.reports'.format(self.package)
    
    def execute(self, args=()):
        """Parse command line arguments, execute the command, and generate
        summary reports.
        
        Returns:
            The return code.
        """
        options = self.parse_args(args)
        retcode, summary = self.run_command(options)
        if retcode == 0 and options.report_file:
            logging.getLogger().debug(
                "Writing report to %s", options.report_file)
            self.generate_reports(summary, options)
        else:
            logging.getLogger().debug("Not generating report file")
        return retcode, summary
    
    def get_command_parser_class(self):
        """Returns the CommandParser class within the cli module.
        """
        mod = import_module(self.cli_module)
        return mod.CommandParser
    
    @property
    def usage(self):
        """Returns the command's usage string.
        """
        return self.get_command_parser_class().usage
    
    @property
    def description(self):
        """Returns the command's description string.
        """
        return self.get_command_parser_class().description
    
    def get_help(self, fmt="* {name}: {description}", wrap=80, indent=2):
        """Returns a string to include in the command help.
        """
        helpstr = fmt.format(
            name=self.name, description=self.description.strip())
        if wrap:
            helpstr = "\n".join(textwrap.wrap(
                re.sub(r'\s+', ' ', helpstr), wrap,
                subsequent_indent=' ' * indent))
        return helpstr
    
    def parse_args(self, args):
        """Parse the command line options.
        
        Returns:
            A Namespace-like object.
        """
        parser_class = self.get_command_parser_class()
        parser = parser_class()
        return parser.parse(args)
    
    def get_command_runner_class(self):
        """Returns the CommandRunner class within the top-level module.
        """
        mod = import_module(self.package)
        return mod.CommandRunner
    
    def run_command(self, options):
        """Run the command.
        
        Args:
            options: A Namespace-like object.
        
        Returns:
            Tuple (retcode, summary).
        """
        runner_class = self.get_command_runner_class()
        runner = runner_class(options)
        return runner.run()
    
    def get_report_generator_class(self):
        """Returns the ReportGenerator class within the reports module.
        """
        mod = import_module(self.report_module)
        return mod.ReportGenerator
    
    def generate_reports(self, summary, options):
        """Generate reports.
        
        Args:
            summary: The summary dict.
            options: Command-line options.
            report_file: The report file name/prefix.
            report_formats: A list of formats.
        """
        generator_class = self.get_report_generator_class()
        generator = generator_class(options)
        generator.generate_reports(summary)

COMMANDS = dict(
    (name, Command(name))
    for _, name, ispkg in walk_packages([os.path.dirname(__file__)])
    if ispkg)

def get_command(name):
    """Returns the Command object for the specified command name.
    """
    if name not in COMMANDS:
        raise ValueError("Invalid command: {}".format(name))
    return COMMANDS[name]

def iter_commands():
    """Iterate over commands, ordered by command name.
    """
    for name in sorted(COMMANDS.keys()):
        yield COMMANDS[name]

def execute_cli(args=()):
    """Executes the Atropos command-line interface.
    
    The first argument is expected to be the command name. If not (i.e. args is
    empty or the first argument starts with a '-'), the 'trim' command is
    assumed. If the first argument is '-h' or '--help', the command-level help
    is printed.
        
    Args:
        args: Command-line arguments.
    
    Returns:
        The return code.
    """
    if len(args) == 0 or args[0] in ('-h', '--help'):
        print_subcommands()
        return 2
    
    config_args = None
    
    if args[0] == '--config':
        with open(args[1], 'rt') as config_file:
            config_args = list(
                token
                for line in config_file
                for token in line.rstrip().split())
        args = args[2:]
    
    def parse_command(args):
        if not args or args[0][0] == '-':
            return ('trim', args)
        else:
            return (args[0], args[1:])
    
    if len(args) == 0:
        command_name, args = parse_command(config_args)
    else:
        command_name, args = parse_command(args)
        if config_args:
            args = config_args + args
    
    try:
        command = get_command(command_name)
        retcode, summary = command.execute(args)
        if 'exception' in summary:
            logging.getLogger().error(
                "Error executing command %s", command_name, 
                exc_info=summary['exception']['details'])
        return retcode
    except Exception as err:
        logging.getLogger().error(
            "Error executing command: %s", command_name, exc_info=err)
        return 2

def print_subcommands():
    """Prints usage message listing the available subcommands.
    """
    print(__doc__.format(__version__, "\n".join(
        command.get_help() for command in iter_commands())))



"""Common classes/functions used in commands.
"""
from collections import Sequence
import copy
import platform
import sys
from atropos import __version__, AtroposError
from atropos.adapters import AdapterCache
from atropos.io.seqio import open_reader, sra_reader
from atropos.util import MergingDict, Const, Summarizable, Timing

class Pipeline(object):
    """Base class for analysis pipelines.
    """
    def __init__(self):
        self.record_counts = {}
        self.bp_counts = {}
    
    def __call__(self, command_runner, raise_on_error=False, **kwargs):
        self.start(**kwargs)
        try:
            for batch in command_runner.iterator():
                self.process_batch(batch)
        except Exception as err:
            if raise_on_error:
                raise
            else:
                command_runner.summary['exception'] = dict(
                    message=str(err),
                    details=sys.exc_info())
        finally:
            self.finish(command_runner.summary, **kwargs)
    
    def start(self, **kwargs):
        """Start the pipeline.
        """
        pass
    
    def process_batch(self, batch):
        """Run the pipeline on a batch of records.
        
        Args:
            batch: A batch of reads. A batch has the format
            ({batch_metadata}, [records]).
        """
        batch_meta, records = batch
        context = batch_meta.copy()
        
        if not context['source'] in self.record_counts:
            self.record_counts[context['source']] = 0
        self.record_counts[context['source']] += context['size']
        
        if not context['source'] in self.bp_counts:
            self.bp_counts[context['source']] = [0, 0]
        context['bp'] = self.bp_counts[context['source']]
        
        self.add_to_context(context)
        self.handle_records(context, records)
    
    def add_to_context(self, context):
        """Add items to the batch context.
        """
        pass
    
    def handle_records(self, context, records):
        """Handle a sequence of records.
        
        Args:
            context: The pipeline context (dict).
            records: The sequence of records.
        """
        for idx, record in enumerate(records):
            try:
                self.handle_record(context, record)
            except Exception as err:
                raise AtroposError(
                    "An error occurred at record {} of batch {}".format(
                    idx, context['index'])) from err
    
    def handle_record(self, context, record):
        """Handle a single record.
        
        Args:
            context: The pipeline context (dict).
            record: The record.
        """
        raise NotImplementedError()
    
    def handle_reads(self, context, read1, read2=None):
        """Handle a read or read-pair.
        
        Args:
            context: The pipeline context (dict).
            read1, read2: The read pair; read2 will be None for single-end data.
        """
        raise NotImplementedError()
    
    def finish(self, summary, **kwargs):
        """Finish the pipeline, including adding information to the summary.
        
        Args:
            summary: Summary dict to update.
        """
        total_bp_counts = tuple(sum(b) for b in zip(*self.bp_counts.values()))
        summary.update(
            record_counts=self.record_counts,
            total_record_count=sum(self.record_counts.values()),
            bp_counts=self.bp_counts,
            total_bp_counts=total_bp_counts,
            sum_total_bp_count=sum(total_bp_counts))

class SingleEndPipelineMixin(object):
    """Mixin for pipelines that implements `handle_record` for single-end data.
    """
    def handle_record(self, context, record):
        context['bp'][0] += len(record)
        return self.handle_reads(context, record)

class PairedEndPipelineMixin(object):
    """Mixin for pipelines that implements `handle_record` for paired-end data.
    """
    def handle_record(self, context, record):
        read1, read2 = record
        bps = context['bp']
        bps[0] += len(read1.sequence)
        bps[1] += len(read2.sequence)
        return self.handle_reads(context, read1, read2)

class Summary(MergingDict):
    """Contains summary information.
    """
    @property
    def has_exception(self):
        return 'exception' in self
    
    def finish(self):
        """Replaces Summarizable members with their summaries, and computes
        some aggregate values.
        """
        self._post_process_dict(self)
    
    def _post_process_dict(self, dict_val):
        if dict_val is None:
            return
        for key, value in tuple(dict_val.items()):
            if value is None:
                continue
            if isinstance(value, Summarizable):
                dict_val[key] = value = value.summarize()
            if isinstance(value, dict):
                self._post_process_dict(value)
            elif (
                    isinstance(value, Sequence) and
                    len(value) > 0 and
                    all(
                        val is None or isinstance(val, dict)
                        for val in value)):
                for val in value:
                    self._post_process_dict(val)
            else:
                if isinstance(value, Const):
                    dict_val[key] = value = value.value
                self._post_process_other(dict_val, key, value)
    
    def _post_process_other(self, parent, key, value):
        pass

class BaseCommandRunner(object):
    """Base class for command executors.
    
    Args:
        options: Command-line options.
    """
    def __init__(self, options, summary_class=Summary):
        self.options = options
        self.summary = summary_class()
        self.timing = Timing()
        self.return_code = None
        self.size = options.batch_size or 1000
        self.batches = 0
        self.done = False
        self._empty_batch = [None] * self.size
        self._progress_options = None
        
        if options.sra_reader:
            self.reader = reader = sra_reader(
                reader=options.sra_reader, 
                quality_base=options.quality_base, 
                colorspace=options.colorspace, 
                input_read=options.input_read,
                alphabet=options.alphabet)
            options.sra_reader = None
        else:
            interleaved = bool(options.interleaved_input)
            input1 = options.interleaved_input if interleaved else options.input1
            input2 = qualfile = None
            if options.paired and not interleaved:
                input2 = options.input2
            else:
                qualfile = options.input2
            self.reader = reader = open_reader(
                file1=input1, file2=input2, file_format=options.format, 
                qualfile=qualfile, quality_base=options.quality_base, 
                colorspace=options.colorspace, interleaved=interleaved, 
                input_read=options.input_read, alphabet=options.alphabet)
        
        # Wrap reader in subsampler
        if options.subsample:
            import random
            if options.subsample_seed:
                random.seed(options.subsample_seed)
            
            def subsample(reader, frac):
                """Generator that yields a random subsample of records.
                
                Args:
                    reader: The reader from which to sample.
                    frac: The fraction of records to yield.
                """
                for reads in reader:
                    if random.random() < frac:
                        yield reads
            
            reader = subsample(reader, options.subsample)
        
        self.iterable = enumerate(reader, 1)
        
        if options.progress:
            self._progress_options = (
                options.progress, self.size, self.max_reads,
                options.counter_magnitude)
        
        self.init_summary()
    
    def __getattr__(self, name):
        if hasattr(self.reader, name):
            return getattr(self.reader, name)
        elif hasattr(self.options, name):
            return getattr(self.options, name)
        else:
            raise ValueError("Unknown attribute: {}".format(name))
    
    def iterator(self):
        """Returns an iterator (an object with the __iter__ method) over
        input batches. BaseCommandRunner is itself an iterable, and will be
        wrapped with a progress bar if _progress_options is set.
        """
        if self._progress_options:
            # Wrap iterator in progress bar
            from atropos.io.progress import create_progress_reader
            itr = create_progress_reader(self, *self._progress_options)
            # itr may be none if there are no progress bar libraries available
            if itr is not None:
                return itr
        return self

    def __iter__(self):
        return self
    
    def __next__(self):
        if self.done:
            raise StopIteration()
        
        try:
            read_index, record = next(self.iterable)
        except:
            self.finish()
            raise
        
        batch = copy.copy(self._empty_batch)
        batch[0] = record
        batch_index = 1
        max_size = self.size
        if self.max_reads:
            max_size = min(max_size, self.max_reads - read_index + 1)
        
        while batch_index < max_size:
            try:
                read_index, record = next(self.iterable)
                batch[batch_index] = record
                batch_index += 1
            except StopIteration:
                self.finish()
                break
            except:
                self.finish()
                raise
        
        if self.max_reads and read_index >= self.max_reads:
            self.finish()
        
        self.batches += 1
        
        batch_meta = dict(
            index=self.batches,
            # TODO: When multi-file input is supported, 'source' will need to
            # be the index of the current file/pair from which records are
            # being read.
            source=0,
            size=batch_index)
        
        if batch_index == self.size:
            return (batch_meta, batch)
        else:
            return (batch_meta, batch[0:batch_index])
    
    def init_summary(self):
        """Initialize the summary dict with general information.
        """
        self.summary['program'] = 'Atropos'
        self.summary['version'] = __version__
        self.summary['python'] = platform.python_version()
        self.summary['command'] = self.name
        self.summary['options'] = self.options.__dict__.copy()
        self.summary['timing'] = self.timing
        self.summary['sample_id'] = self.options.sample_id
        self.summary['input'] = self.reader.summarize()
        self.summary['input'].update(
            batch_size=self.size,
            max_reads=self.max_reads,
            batches=self.batches)
    
    def run(self):
        """Run the command, wrapping it in a Timing, catching any exceptions,
        and finally closing the reader.
        
        Returns:
            The tuple (retcode, summary).
        """
        with self.timing:
            try:
                self.return_code = self()
            except Exception as err: # pylint: disable=broad-except
                self.summary['exception'] = dict(
                    message=str(err),
                    details=sys.exc_info())
                self.return_code = 1
            finally:
                self.finish()
        
        return (self.return_code, self.summary)
    
    def __call__(self):
        """Execute the command. Must be implemented within the command
        module.
        
        Returns:
            The return code.
        """
        raise NotImplementedError()
    
    def finish(self):
        """Finish the command.
        """
        # Close the underlying reader.
        if not self.done:
            self.done = True
            self.reader.close()
        self.summary.finish()
    
    def load_known_adapters(self):
        """Load known adapters based on setting in command-line options.
        
        Args:
            options: Command-line options.
        """
        cache_file = None
        if self.options.cache_adapters:
            cache_file = self.options.adapter_cache_file
        adapter_cache = AdapterCache(cache_file)
        if adapter_cache.empty and self.options.default_adapters:
            adapter_cache.load_default()
        if self.options.known_adapter:
            for known in self.options.known_adapter:
                name, seq = known.split('=')
                adapter_cache.add(name, seq)
        if self.options.known_adapters_file:
            for known_file in self.options.known_adapters_file:
                adapter_cache.load_from_url(known_file)
        if self.options.cache_adapters:
            adapter_cache.save()
        return adapter_cache



# coding: utf-8
"""Classes for writing and filtering of processed reads.

A Filter is a callable that has the read as its only argument. If it is called,
it returns True if the read should be filtered (discarded), and False if not.

To be used, a filter needs to be wrapped in one of the Wrapper classes. To determine
what happens to a read, a list of Wrappers with different filters is created and
each Wrapper is called in turn until one returns True. The main program will
determine whether and where to write the read(s) based on whether it was rejected
by a filter, and which one.
"""
from collections import OrderedDict

# Constants used when returning from a Filter’s __call__ method to improve
# readability (it is unintuitive that "return True" means "discard the read").
DISCARD = True
KEEP = False

class FilterWrapper(object):
    """Base wrapper around a filter.
    """
    def __init__(self, f):
        self.filtered = 0
        self.filter = f
    
    def __call__(self, read1, read2=None):
        """Call the filter function.
        
        Returns:
            DISCARD if the filter function returns True, else KEEP
        """
        if self._filter(read1, read2):
            self.filtered += 1
            return DISCARD
        return KEEP
    
    def _filter(self, read1, read2=None):
        """Call the filter function.
        """
        raise NotImplementedError()
    
    @property
    def name(self):
        """The filter name.
        """
        if hasattr(self.filter, 'name'):
            return self.filter.name
        else:
            return self.filter.__class__.__name__
    
    def summarize(self):
        """Returns a summary dict.
        """
        return dict(records_filtered=self.filtered)

class SingleWrapper(FilterWrapper):
    """This is for single-end reads and for paired-end reads, using the 'legacy'
    filtering mode (backwards compatibility). That is, if the first read matches
    the filtering criteria, the pair is discarded. The second read is not
    inspected.
    """
    def _filter(self, read1, read2=None):
        return self.filter(read1)

class PairedWrapper(FilterWrapper):
    """This is for paired-end reads, using the 'new-style' filtering where both
    reads are inspected. That is, the entire pair is discarded if at least 1 or
    2 of the reads match the filtering criteria.
    
    Args:
        min_affected: values 1 and 2 are allowed.
            1 means: the pair is discarded if any read matches
            2 means: the pair is discarded if both reads match
    """
    def __init__(self, f, min_affected=1):
        super().__init__(f)
        if min_affected not in (1, 2):
            raise ValueError("min_affected must be 1 or 2")
        self.min_affected = min_affected

    def _filter(self, read1, read2):
        failures = 0
        if self.filter(read1):
            failures += 1
        if (
                (self.min_affected - failures == 1) and
                (read2 is None or self.filter(read2))):
            failures += 1
        return failures >= self.min_affected

class FilterFactory(object):
    """Factor that creates filters and wraps them in the appropriate
    (single-end or paired-end) wrapper.
    """
    def __init__(self, paired, min_affected):
        self.paired = paired
        self.min_affected = min_affected
    
    def __call__(self, filter_type, *args, **kwargs):
        """Create a wrapped filter of the specified type.
        """
        fltr = filter_type(*args, **kwargs)
        if self.paired == "both":
            return PairedWrapper(fltr, self.min_affected)
        else:
            return SingleWrapper(fltr)

class MergedReadFilter(object):
    """Returns True if the read was merged.
    """
    def __call__(self, read):
        return read.merged

class TooShortReadFilter(object):
    """Returns True if the read sequence is shorter than `minimum_length`.
    """
    name = "too_short"
    
    def __init__(self, minimum_length):
        self.minimum_length = minimum_length
    
    def __call__(self, read):
        return len(read) < self.minimum_length

class TooLongReadFilter(object):
    """Returns True if the read sequence is longer than `maximum_length`.
    """
    name = "too_long"
    
    def __init__(self, maximum_length):
        self.maximum_length = maximum_length
    
    def __call__(self, read):
        return len(read) > self.maximum_length

class NContentFilter(object):
    """Discards a reads that has a number of 'N's over a given threshold. It
    handles both raw counts of Ns as well as proportions. Note, for raw counts,
    it is a greater than comparison, so a cutoff of '1' will keep reads with a
    single N in it.
    
    Args:
        count: If it is below 1.0, it will be considered a proportion, and
            above and equal to 1 will be considered as discarding reads with a
            number of N's greater than this cutoff.
    """
    name = "too_many_n"
    
    def __init__(self, count):
        assert count >= 0
        self.is_proportion = count < 1.0
        self.cutoff = count

    def __call__(self, read):
        """Return True when the read should be discarded.
        """
        n_count = read.sequence.lower().count('n')
        if self.is_proportion:
            if len(read) == 0:
                return False
            return n_count / len(read) > self.cutoff
        else:
            return n_count > self.cutoff

class UntrimmedFilter(object):
    """Returns True if read is untrimmed.
    """
    def __call__(self, read):
        return read.match is None

class TrimmedFilter(object):
    """Returns True if read is trimmed.
    """
    def __call__(self, read):
        return read.match is not None

class NoFilter(object):
    """Always returns False.
    """
    name = "NoFilter"
    
    def __call__(self, read):
        return False

class Filters(object):
    """Manages multiple filters.
    
    Args:
        filter_factory: The :class:`FilterFactory` to use for creating new
            filters.
    """
    def __init__(self, filter_factory):
        self.filters = OrderedDict()
        self.filter_factory = filter_factory
    
    def add_filter(self, filter_type, *args, **kwargs):
        """Add a filter of the specified type.
        """
        self.filters[filter_type] = self.filter_factory(
            filter_type, *args, **kwargs)
    
    def filter(self, read1, read2=None):
        """Execute filters in the order they were added until one returns True.
        
        Args:
            read1, read2: The read(s) to filter.
        
        Returns:
            The type of the first filter that returned True, or :class:NoFilter
            if none of the filters returned True.
        """
        dest = NoFilter
        for filter_type, fltr in self.filters.items():
            if fltr(read1, read2):
                dest = filter_type
                # Stop writing as soon as one of the filters was successful.
                break
        return dest
    
    def __contains__(self, filter_type):
        return filter_type in self.filters
    
    def __getitem__(self, filter_type):
        return self.filters[filter_type]
    
    def summarize(self):
        """Returns a summary dict.
        """
        return dict(
            (f.name, f.summarize())
            for f in self.filters.values())



# coding: utf-8
from pytest import raises
from atropos.adapters import (
    Adapter, Match, ColorspaceAdapter, FRONT, BACK, parse_braces, LinkedAdapter
)
from atropos.io.seqio import Sequence


def test_issue_52():
    adapter = Adapter(
        sequence='GAACTCCAGTCACNNNNN',
        where=BACK,
        max_error_rate=0.12,
        min_overlap=5,
        read_wildcards=False,
        adapter_wildcards=True,
    )
    read = Sequence(name="abc", sequence='CCCCAGAACTACAGTCCCGGC')
    am = Match(
        astart=0,
        astop=17,
        rstart=5,
        rstop=21,
        matches=15,
        errors=2,
        front=None,
        adapter=adapter,
        read=read,
    )
    assert am.wildcards() == 'GGC'
    """
    The result above should actually be 'CGGC' since the correct
    alignment is this one:

    adapter         GAACTCCAGTCACNNNNN
    mismatches           X     X
    read       CCCCAGAACTACAGTC-CCGGC

    Since we do not keep the alignment, guessing 'GGC' is the best we
    can currently do.
    """


def test_issue_80():
    # This issue turned out to not be an actual issue with the alignment
    # algorithm. The following alignment is found because it has more matches
    # than the 'obvious' one:
    #
    # TCGTATGCCGTCTTC
    # =========X==XX=
    # TCGTATGCCCTC--C
    #
    # This is correct, albeit a little surprising, since an alignment without
    # indels would have only two errors.
    adapter = Adapter(
        sequence="TCGTATGCCGTCTTC",
        where=BACK,
        max_error_rate=0.2,
        min_overlap=3,
        read_wildcards=False,
        adapter_wildcards=False,
    )
    read = Sequence(name="seq2", sequence="TCGTATGCCCTCC")
    result = adapter.match_to(read)
    assert read.original_length == 13, result
    assert result.errors == 3, result
    assert result.astart == 0, result
    assert result.astop == 15, result


def test_str():
    a = Adapter('ACGT', where=BACK, max_error_rate=0.1)
    str(a)
    str(a.match_to(Sequence(name='seq', sequence='TTACGT')))
    ca = ColorspaceAdapter('0123', where=BACK, max_error_rate=0.1)
    str(ca)


def test_color():
    with raises(ValueError):
        ColorspaceAdapter('0123', where=FRONT, max_error_rate=0.1)


def test_parse_braces():
    assert parse_braces('') == ''
    assert parse_braces('A') == 'A'
    assert parse_braces('A{0}') == ''
    assert parse_braces('A{1}') == 'A'
    assert parse_braces('A{2}') == 'AA'
    assert parse_braces('A{2}C') == 'AAC'
    assert parse_braces('ACGTN{3}TGACCC') == 'ACGTNNNTGACCC'
    assert parse_braces('ACGTN{10}TGACCC') == 'ACGTNNNNNNNNNNTGACCC'
    assert parse_braces('ACGTN{3}TGA{4}CCC') == 'ACGTNNNTGAAAACCC'
    assert parse_braces('ACGTN{0}TGA{4}CCC') == 'ACGTTGAAAACCC'


def test_parse_braces_fail():
    for expression in [
        '{',
        '}',
        '{}',
        '{5',
        '{1}',
        'A{-7}',
        'A{',
        'A{1',
        'N{7',
        'AN{7',
        'A{4{}',
        'A{4}{3}',
        'A{b}',
        'A{6X}',
        'A{X6}',
    ]:
        with raises(ValueError):
            parse_braces(expression)


def test_linked_adapter():
    linked_adapter = LinkedAdapter('AAAA', 'TTTT')
    sequence = Sequence(name='seq', sequence='AAAACCCCCTTTT')
    match = linked_adapter.match_to(sequence)
    trimmed = linked_adapter.trimmed(match)
    assert trimmed.name == 'seq'
    assert trimmed.sequence == 'CCCCC'


def test_random_match_probabilities():
    a = Adapter('A', BACK)
    rmp = a.random_match_probabilities()
    assert rmp == [1.0, 0.25]
    a = Adapter('AC', BACK, gc_content=0.4)
    rmp = a.random_match_probabilities()
    assert rmp == [1.0, 0.3, 0.06]



"""Interface to report generation.
"""
import importlib
import os
from atropos.io import STDOUT, STDERR, open_output
from atropos.io.seqio import PAIRED

SERIALIZERS = dict(
    json='t',
    yaml='t',
    pickle='b'
)

class BaseReportGenerator(object):
    """Base class for command report generators.
    
    Args:
        options: Command-line options.
    """
    def __init__(self, options):
        report_file = options.report_file
        report_formats = options.report_formats
        if report_file in (STDOUT, STDERR):
            self.report_formats = report_formats or ('txt',)
            self.report_files = (report_file,) * len(self.report_formats)
        else:
            file_parts = os.path.splitext(report_file)
            self.report_formats = (
                report_formats or
                (file_parts[1][1:] if file_parts[1] else 'txt',))
            if len(self.report_formats) == 1:
                self.report_files = (report_file,)
            else:
                self.report_files = (
                    '{}.{}'.format(report_file, fmt)
                    for fmt in self.report_formats)
        self.report_args = tuple(
            self.get_report_args(fmt, options)
            for fmt in self.report_formats)
    
    def get_report_args(self, fmt, options):
        """Returns report-specific options dict.
        """
        return {}
    
    def generate_reports(self, summary):
        """Generate report(s) from a summary.
        
        Args:
            summary: The summary dict.
            report_file: File name (if generating a single report) or file name
                prefix.
            report_formats: Sequence of formats to generate.
            kwargs: Additional arguments to pass to report generator
                function(s).
        """
        self.add_derived_data(summary)
        
        for fmt, outfile, kwargs in zip(
                self.report_formats, self.report_files, self.report_args):
            if fmt in SERIALIZERS:
                mode = SERIALIZERS[fmt]
                self.serialize(summary, fmt, mode, outfile, **kwargs)
            else:
                self.generate_text_report(fmt, summary, outfile, **kwargs)
    
    def add_derived_data(self, summary):
        """Get some additional fields that are useful in the report.
        """
        derived = {}
        derived['mean_sequence_lengths'] = tuple(
            None if bp is None else bp / summary['total_record_count']
            for bp in summary['total_bp_counts'])
        
        inp = summary['input']
        fmt = inp['file_format']
        if inp['input_read'] == PAIRED:
            fmt += ', Paired'
        else:
            fmt += ', Read {}'.format(inp['input_read'])
        if inp['colorspace']:
            fmt += ', Colorspace'
        if inp['interleaved']:
            fmt += ', Interleaved'
        if inp['delivers_qualities']:
            fmt += ', w/ Qualities'
        else:
            fmt += ', w/o Qualities'
        derived['input_format'] = fmt
        
        summary['derived'] = derived
    
    def serialize(self, obj, fmt, mode, outfile, **kwargs):
        """Serialize a summary dict to a file.
        
        Args:
            obj: The summary dict.
            fmt: The serialization format (e.g. json, yaml).
            mode: The file mode (b=binary, t=text).
            outfile: The output file.
            kwargs: Additional arguments to pass to the `dump` method.
        """
        mod = importlib.import_module(fmt)
        with open_output(outfile, 'w' + mode) as stream:
            mod.dump(obj, stream, **kwargs)
    
    def generate_text_report(self, fmt, summary, outfile, **kwargs):
        """Generate a text report. By default, calls generate_from_template.
        """
        self.generate_from_template(fmt, summary, outfile, **kwargs)
    
    def generate_from_template(
            self, fmt, summary, outfile, template_name=None,
            template_paths=None, template_globals=None):
        """Generate a report using Jinja2.
        
        Args:
            fmt: Report format. Must match the file extension on a discoverable
                template.
            summary: The summary dict.
            outfile: The output file name/prefix.
            template_name: A template name to use, rather than auto-discover.
            template_paths: Sequence of paths to search for templates.
            template_globals: Dict of additional globals to add to the template
                environment.
        """
        import jinja2
        
        if not template_name:
            template_name = 'template.{}'.format(fmt)
        if not template_paths:
            template_paths = []
        if hasattr(self, 'template_path'):
            template_paths.append(self.template_path)
        
        # Load the report template
        try:
            env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(template_paths))
            if template_globals:
                env.globals.update(template_globals)
            template = env.get_template(template_name)
        except:
            raise IOError(
                "Could not load template file '{}'".format(template_name))

        # Render the template
        report_output = template.render(summary=summary)
        
        # Write to output
        is_path = isinstance(outfile, str)
        if is_path:
            outfile = open(outfile, 'wt', encoding='utf-8')
        else:
            report_output = report_output.encode('utf-8')
        
        try:
            print(report_output, file=outfile)
        except IOError as err:
            raise IOError(
                "Could not print report to '{}' - {}".format(outfile, err))
        finally:
            if is_path:
                outfile.close()

def prettyprint_summary(summary, outfile='summary.dump.txt'):
    """Pretty-print the summary to a file. Mostly used for debugging.
    """
    from pprint import pprint
    with open(outfile, 'w') as out:
        pprint(summary, out)
